"""
Integration test for running parallel jobs
This is kept separate from unittest because of
undefined behavior when running unittests on thread leads to
InactiveRPC error even though server is running and available.
"""
import concurrent.futures
import time

import numpy as np

from eqc_direct.client import EqcClient
from eqc_direct.utils import convert_hamiltonian_to_poly_format


def run_eqc_job(n):
    print("N input:", n)
    h = np.random.normal(size=(n, 1)).astype(np.float32)
    J = np.random.normal(size=(n, n)).astype(np.float32)
    ham_mat = np.hstack((h, J + J.T))
    print(ham_mat)
    poly_indices, poly_coefficients = convert_hamiltonian_to_poly_format(
        linear_terms=np.squeeze(ham_mat[:, 0]),
        quadratic_terms=ham_mat[:, 1:],
    )
    start = time.time()
    eqc_client = EqcClient()
    lock_id, start_ts, end_ts = eqc_client.wait_for_lock()
    print("Lock:", lock_id)
    resp = eqc_client.solve_sum_constrained(
        poly_coefficients=poly_coefficients, poly_indices=poly_indices, lock_id=lock_id
    )
    eqc_client.release_lock(lock_id=lock_id)
    print("Time Job:", time.time() - start)
    print("Time_Queue:", (end_ts - start_ts) / (10**9))
    print("Time_Run:", (resp["end_job_ts"] - resp["start_job_ts"]) / 10**9)
    print("Matches input:", n == len(resp["solution"][0]))
    print("N returned", len(resp["solution"][0]))
    return n == len(resp["solution"][0])


if __name__ == "__main__":
    # max problem size is 949
    n_jobs = 10
    start = time.time()
    prob_sizes = np.random.randint(20, 949, size=n_jobs).tolist()
    with concurrent.futures.ThreadPoolExecutor(max_workers=3) as process_pool:
        x = list(process_pool.map(run_eqc_job, prob_sizes))
    print("Matched return:", x)
    print("TOTAL TIME:", time.time() - start)



import sys
import time
from multiprocessing import Pool
from eqc_direct.eqc_client import EqcClient
import numpy as np


def check_status():
    eqc_client = EqcClient()
    resp = eqc_client.system_status()
    print(resp)
    print("Float 32 type", ham_mat_32.dtype)


if __name__ == "__main__":
    check_status()



import sys
import time
import numpy as np
from eqc_direct.client import EqcClient
from eqc_direct.utils import convert_hamiltonian_to_poly_format


def run_single():
    ham_mat = np.round(
        np.loadtxt("tam_200_example_2.csv", delimiter=",", dtype=np.float32)
    )
    poly_idx, poly_coef = convert_hamiltonian_to_poly_format(
        linear_terms=ham_mat[:, 0], quadratic_terms=ham_mat[:, 1:]
    )
    print(ham_mat.shape)
    start = time.time()
    eqc_client = EqcClient()
    lock_id, start_ts, end_ts = eqc_client.wait_for_lock()
    print("Lock:", lock_id)
    resp = eqc_client.process_job(
        poly_coefficients=poly_coefs,
        poly_indices=poly_idx,
        sum_constraint=100,
        relaxation_schedule=2,
        solution_precision=1,
    )

    eqc_client.release_lock(lock_id=lock_id)
    print(resp)
    print("Execution time:", time.time() - start)


if __name__ == "__main__":
    run_single()



import os
import threading
import time
import unittest
import warnings

import numpy as np
from grpc_server_cls import ServerForTest
from eqc_direct.client import EqcClient
from eqc_direct.server_sim import GrpcServer
from eqc_direct.utils import JobCodes, LockCheckStatus, LockManageStatus


class TestEqcClient(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.test_ip_addr = "localhost"
        # only instantiate server if running on localhost
        cls.server = ServerForTest(
            ip_address=cls.test_ip_addr,
            port="50051",
            private_key="private.pem",
            cert_file="certificate.pem",
        )
        cls.server.start()
        cls.eqc_client = EqcClient(
            ip_address=cls.test_ip_addr, cert_file="certificate.pem"
        )
        cls.poly_indices = np.array(
            [
                [0, 0, 0, 3],
                [0, 0, 2, 2],
                [0, 0, 2, 3],
                [0, 1, 2, 3],
                [0, 1, 1, 1],
                [0, 2, 2, 3],
                [3, 3, 3, 3],
                [1, 1, 2, 3],
            ],
        )
        cls.poly_coefficients = np.array([1, 2, 3, 4, 5, 6, 7, 8], dtype=np.float32)
        cls.num_variables = 3
        cls.num_samples = 7

    @classmethod
    def tearDownClass(cls):
        if cls.test_ip_addr == "localhost":
            cls.server.stop()

    def test_check_lock(self):
        try:
            lock_av = self.eqc_client.check_lock()
            print("lock_av", lock_av)
            self.assertDictEqual(lock_av, LockCheckStatus.AVAILABLE)
            lock_id = self.eqc_client.acquire_lock()["lock_id"]
            lock_unv = self.eqc_client.check_lock()
            print("lock_unv", lock_unv)
            self.assertDictEqual(lock_unv, LockCheckStatus.UNAVAILABLE)
            lock_user = self.eqc_client.check_lock(lock_id=lock_id)
            self.assertDictEqual(lock_user, LockCheckStatus.USER_LOCKED)
        finally:
            self.eqc_client.release_lock(lock_id=lock_id)

    def test_basic_locking(self):
        try:
            print("Test basic locking")
            succ_lock_resp = self.eqc_client.acquire_lock()
            lock_id = succ_lock_resp["lock_id"]
            self.assertIsInstance(lock_id, str)
            self.assertEqual(36, len(lock_id))
            lock_fail = self.eqc_client.acquire_lock()
            self.assertEqual(lock_fail["lock_id"], "")

            self.assertEqual(lock_fail["status_code"], 2)
            self.assertEqual(
                lock_fail["status_desc"],
                "Lock currently in use unable to perform operation",
            )
        finally:
            release_out = self.eqc_client.release_lock(lock_id=lock_id)
            self.assertEqual(release_out["status_code"], 0)
            self.assertEqual(release_out["status_desc"], "Success")

    def test_wait_for_lock(self):
        print("Test wait for lock")
        try:
            lock_id, start_ts, end_ts = self.eqc_client.wait_for_lock()
            self.assertIsInstance(lock_id, str)
            self.assertEqual(36, len(lock_id))
            self.assertTrue(start_ts < end_ts)
            self.assertIsInstance(start_ts, int)
            self.assertIsInstance(end_ts, int)
        finally:
            self.eqc_client.release_lock(lock_id=lock_id)

    def test_stop_running_process(self):
        try:
            print("Stop test")
            lock_id, _, _ = self.eqc_client.wait_for_lock()
            # this isn't a real problem since isn't symmetric
            # just chose one that is large enough to take a significant
            # time to do data transfer so can cancel.
            print("stop lock", lock_id)
            submit_output = self.eqc_client.submit_sum_constrained_job(
                poly_indices=self.poly_indices,
                poly_coefficients=self.poly_coefficients,
                num_variables=self.num_variables,
                relaxation_schedule=1,
                sum_constraint=40,
                lock_id=lock_id,
            )
            print("SUBMTI OUTPUT", submit_output)
            stop_out = self.eqc_client.stop_running_process(lock_id=lock_id)
            self.assertDictEqual(stop_out, LockManageStatus.SUCCESS)
            stop_res = self.eqc_client.fetch_sum_constrained_result(lock_id=lock_id)
            self.assertEqual(stop_res["err_code"], JobCodes.USER_INTERRUPT["err_code"])
            self.assertEqual(stop_res["err_desc"], JobCodes.USER_INTERRUPT["err_desc"])

        finally:
            self.eqc_client.release_lock(lock_id=lock_id)

    def test_system_info(self):
        version_resp = self.eqc_client.system_info()
        print("VERSION", version_resp)
        self.assertIsInstance(version_resp["server_version"], str)
        self.assertIsInstance(version_resp["device_type"], str)

    def test_system_status(self):
        sys_resp = self.eqc_client.system_status()
        self.assertListEqual(list(sys_resp.keys()), ["status_code", "status_desc"])
        self.assertIsInstance(sys_resp["status_code"], int)
        self.assertIsInstance(sys_resp["status_desc"], str)

    def test_solve_sum_constrained_and_fetch_results_batched_samples_non_distilled(
        self,
    ):
        try:
            print("Process job tests with batched samples for non-distilled solution")
            lock_id, _, _ = self.eqc_client.wait_for_lock()
            solve_sum_constrained_output = self.eqc_client.solve_sum_constrained(
                poly_indices=self.poly_indices,
                poly_coefficients=self.poly_coefficients,
                relaxation_schedule=1,
                sum_constraint=40,
                lock_id=lock_id,
                num_samples=self.num_samples,
            )
            self.assertEqual(solve_sum_constrained_output["err_code"], 0)
            job_result = self.eqc_client.fetch_sum_constrained_result(lock_id=lock_id)
            self.assertEqual(job_result["err_code"], 0)
            self.assertEqual(job_result["err_desc"], "Success")
            self.assertEqual(job_result["num_samples"], self.num_samples)
            self.assertEqual(job_result["num_variables"], self.num_variables)
            self.assertTrue(job_result["preprocessing_time"] > 0)
            self.assertIsInstance(job_result["calibration_time"], float)
            self.assertTrue(len(job_result["postprocessing_time"]), self.num_samples)
            self.assertTrue(len(job_result["runtime"]) == self.num_samples)
            self.assertTrue(len(job_result["energy"]) == self.num_samples)
            self.assertTrue(len(job_result["solution"]) == self.num_samples)
            self.assertTrue(len(job_result["distilled_energy"]) == 0)
            self.assertTrue(len(job_result["distilled_solution"]) == 0)
            for k in range(self.num_samples):
                self.assertTrue(job_result["runtime"][k] > 0)
                self.assertIsInstance(job_result["energy"][k], float)
                self.assertIsInstance(job_result["postprocessing_time"][k], float)
                self.assertTrue(len(job_result["solution"][k]), self.num_variables)
        finally:
            self.eqc_client.release_lock(lock_id=lock_id)



import time

from eqc_direct.client import EqcClient


if __name__ == "__main__":
    eqc_client = EqcClient()
    lock_id, _, _ = eqc_client.wait_for_lock()
    start_second_wait = time.time()
    print(eqc_client.wait_for_lock())
    print("Time wait:", time.time() - start_second_wait)



"""
Utilities for running server sim and client
"""
from dataclasses import dataclass
from typing import List, Tuple, TypedDict

import numpy as np

# levels for precision calc
PREC_MIN_RECOMMENDED_LEVELS = 200
PREC_MAX_LEVELS = 10000


class SystemInfoOptional(TypedDict, total=False):
    """Python binding to SystemInfo->VersionOutput proto spec optional fields."""

    device_id: str


class SystemInfo(SystemInfoOptional):
    """Python binding to SystemInfo->VersionOutput proto spec."""

    server_version: str
    device_type: str


@dataclass
class SysStatus:
    """
    Status codes for system paired with their descriptions.
    """

    IDLE = {"status_code": 0, "status_desc": "IDLE"}
    JOB_RUNNING = {"status_code": 1, "status_desc": "JOB_RUNNING"}
    CALIBRATION = {"status_code": 2, "status_desc": "CALIBRATION"}
    HEALTH_CHECK = {"status_code": 3, "status_desc": "HEALTH_CHECK"}
    HARDWARE_FAILURE = {"status_code": [4, 5], "status_desc": "HARDWARE_FAILURE"}


@dataclass
class LockCheckStatus:
    """
    Statuses codes for checking lock status paired with their descriptions
    """

    AVAILABLE = {"status_code": 0, "status_desc": "Lock available"}
    USER_LOCKED = {
        "status_code": 1,
        "status_desc": "lock_id matches current server lock_id",
    }
    UNAVAILABLE = {
        "status_code": 2,
        "status_desc": "Execution lock is in use by another user",
    }


@dataclass
class LockManageStatus:
    """
    Statuses and descriptions for acquiring and releasing lock
    """

    SUCCESS = {"status_code": 0, "status_desc": "Success"}
    MISMATCH = {
        "status_code": 1,
        "status_desc": "lock_id does not match current device lock_id",
    }
    BUSY = {
        "status_code": 2,
        "status_desc": "Lock currently in use unable to perform operation",
    }


@dataclass
class JobCodes:
    """
    Job codes for errors paired with their descriptions
    """

    NORMAL = {"err_code": 0, "err_desc": "Success"}
    INDEX_OUT_OF_RANGE = {
        "err_code": 1,
        "err_desc": (
            "Index in submitted data is out of range for specified "
            "number of variables"
        ),
    }
    COEF_INDEX_MISMATCH = {
        "err_code": 2,
        "err_desc": (
            "Polynomial indices do not match required length for "
            "specified coefficient length"
        ),
    }
    DEVICE_BUSY = {
        "err_code": 3,
        "err_desc": "Device currently processing other request",
    }
    LOCK_MISMATCH = {
        "err_code": 4,
        "err_desc": "lock_id doesn't match current device lock",
    }
    HARDWARE_FAILURE = {
        "err_code": 5,
        "err_desc": "Device failed during execution",
    }
    INVALID_SUM_CONSTRAINT = {
        "err_code": 6,
        "err_desc": (
            "Sum constraint must be greater than or equal to 1 and "
            "less than or equal to 10000"
        ),
    }
    INVALID_RELAXATION_SCHEDULE = {
        "err_code": 7,
        "err_desc": "Parameter relaxation_schedule must be in set {1,2,3,4}",
    }
    USER_INTERRUPT = {
        "err_code": 8,
        "err_desc": "User sent stop signal before result was returned",
    }
    EXCEEDS_MAX_SIZE = {
        "err_code": 9,
        "err_desc": "Exceeds max problem size for device",
    }
    DECREASING_INDEX = {
        "err_code": 10,
        "err_desc": (
            "One of specified polynomial indices is not specified in "
            "non-decreasing order"
        ),
    }
    INVALID_PRECISION = {
        "err_code": 11,
        "err_desc": "The input precision exceeds maximum allowed precision for device",
    }
    NUM_SAMPLES_POSITIVE = {
        "err_code": 12,
        "err_desc": "Input num_samples must be positive.",
    }
    PRECISION_CONSTRAINT_MISMATCH = {
        "err_code": 13,
        "err_desc": "Sum constraint must be divisible by solution_precision",
    }
    PRECISION_NONNEGATIVE = {
        "err_code": 14,
        "err_desc": "Input solution precision cannot be negative",
    }
    DEGREE_POSITIVE = {
        "err_code": 15,
        "err_desc": "Input degree must be greater than 0",
    }
    NUM_VARIABLES_POSITIVE = {
        "err_code": 16,
        "err_desc": "Input num_variables must be greater than 0",
    }
    NUM_LEVELS_NUM_VARS_MISMATCH = {
        "err_code": 17,
        "err_desc": "Length of `num_levels` input must be equal to num_variables",
    }
    NUM_LEVELS_GT_ONE = {
        "err_code": 18,
        "err_desc": "All elements of input `num_levels` must be greater than 1",
    }
    TOTAL_INTEGER_LEVELS = {
        "err_code": 19,
        "err_desc": "Total number of integer levels from input variables exceeds limit",
    }
    INVALID_MEAN_PHOTON_NUMBER = {
        "err_code": 20,
        "err_desc": "Mean photon number if specified must be in range [0.0000667, 0.0066666]",
    }
    INVALID_QUANTUM_FLUCTUATION_COEFFICIENT = {
        "err_code": 21,
        "err_desc": "Quantum fluctuation coefficient if specified must be in range [1, 100]",
    }


def message_to_dict(grpc_message) -> dict:
    """Convert a gRPC message to a dictionary."""
    result = {}
    for descriptor in grpc_message.DESCRIPTOR.fields:
        field = getattr(grpc_message, descriptor.name)

        if descriptor.type == descriptor.TYPE_MESSAGE:
            # Handle repeated message fields with nested repeated fields
            if descriptor.label == descriptor.LABEL_REPEATED:
                if descriptor.message_type and any(
                    subfield.label == subfield.LABEL_REPEATED
                    for subfield in descriptor.message_type.fields
                ):
                    # If the message contains repeated fields, extract only the lists
                    result[descriptor.name] = (
                        [list(item.values) for item in field] if field else []
                    )
                else:
                    # Standard repeated message field
                    result[descriptor.name] = (
                        [message_to_dict(item) for item in field] if field else []
                    )
            else:
                # Singular message field
                result[descriptor.name] = message_to_dict(field) if field else {}
        else:
            # Handle repeated primitive fields
            if descriptor.label == descriptor.LABEL_REPEATED:
                result[descriptor.name] = list(field) if field else []
            else:
                result[descriptor.name] = field
    return result


def convert_hamiltonian_to_poly_format(
    linear_terms: np.ndarray,
    quadratic_terms: np.ndarray,
) -> Tuple[List[List[int]], List[float]]:
    """
    Converts linear terms and quadratic terms of Hamiltonian to polynomial
    index formatting for Dirac device

    :param linear_terms: the linear terms for the Hamiltonian 1D length n array
    :param quadratic_terms: the quadratic coefficients of the Hamiltonian (n by n)
    :return: a tuple with the following members:

        - **poly_indices**: List[List[int]] - polynomial indices in non-decreasing sparse format
        - **poly_coefficients**: List[float] - polynomial coefficients in sparse format

    .. Sparse matrices are much slower with this implementation convert to dense first
    """
    assert len(linear_terms.shape) == 1, "`linear_terms` input must be 1D"
    try:
        m, n = quadratic_terms.shape  # pylint: disable=C0103
        assert m == n, "`quadratic_terms` must be a square matrix"
    except ValueError as exc:
        raise ValueError("`quadratic_terms` must be 2D") from exc

    assert len(linear_terms) == m, (
        "`linear_terms` must match dimension row and "
        "column dimension of `quadratic_terms`"
    )
    poly_coefficients = []
    poly_indices = []
    # multiplying by 2 here is faster than doing it at each entry in the upper
    # triangular portion
    quadratic_terms = quadratic_terms * 2
    diag_idx = np.diag_indices(m)
    quadratic_terms[diag_idx] /= 2
    for i in range(m):
        # add nonzero linear terms to lists
        if linear_terms[i] != 0:
            poly_indices.append([0, i + 1])
            poly_coefficients.append(linear_terms[i])
        for j in range(m):
            # add quad terms to lists
            if quadratic_terms[i, j] and i <= j:
                poly_coefficients.append(quadratic_terms[i, j])
                poly_indices.append([i + 1, j + 1])
    return poly_indices, poly_coefficients


def get_decimal_places(float_num: float) -> int:
    """
    Helper function which gets the number of decimal places for a float,
    excluding trailing zeros.

    :param float_num: float input for which decimal places will be found
    :return: a non-negative integer representing the number of decimal places
    """
    try:
        # Split the number into integer and fractional parts
        _, fractional_part = str(float_num).split(".")
        # Strip trailing zeros from the fractional part
        fractional_part = fractional_part.rstrip("0")
        decimal_places = len(fractional_part)
    except ValueError:
        # No fractional part means 0 decimal places
        decimal_places = 0
    return decimal_places



import time

import numpy as np

from eqc_direct.client import EqcClient
from eqc_direct.utils import convert_hamiltonian_to_poly_format


def run_stop_job(n):
    print("N size:", n)
    h = np.random.normal(size=(n, 1))
    J = np.random.normal(size=(n, n))
    ham_mat = np.hstack((h, J + J.T)).astype(np.float32)
    poly_indices, poly_coefficients = convert_hamiltonian_to_poly_format(
        linear_terms=np.squeeze(ham_mat[:, 0]),
        quadratic_terms=ham_mat[:, 1:],
    )
    eqc_client = EqcClient()
    lock_id, _, _ = eqc_client.wait_for_lock()
    print("Lock:", lock_id)
    print("Submit Job")
    resp = eqc_client.submit_sum_constrained_job(
        poly_coefficients=poly_coefficients,
        poly_indices=poly_indices,
        sum_constraint=100,
        lock_id=lock_id,
    )
    print("Resp Job", resp)
    try:
        stop_message = eqc_client.stop_running_process(lock_id=lock_id)
        print("stop_message", stop_message)
    finally:
        lock_status = eqc_client.release_lock(lock_id=lock_id)
    print("Lock status", lock_status)
    return stop_message


if __name__ == "__main__":
    n = 900
    x = run_stop_job(n=n)

    print("Matched return:", x)



import unittest
import numpy as np
from eqc_direct.utils import (
    convert_hamiltonian_to_poly_format,
    get_decimal_places,
)


class TestUtils(unittest.TestCase):
    def test_convert_hamiltonian_to_poly_format(self):
        n = 5
        lin_terms = np.random.normal(size=n)
        quad_terms = np.random.normal(size=(n, n))
        quad_terms = quad_terms.T + quad_terms
        indices, coefs = convert_hamiltonian_to_poly_format(
            linear_terms=lin_terms, quadratic_terms=quad_terms
        )

        n_terms = (n**2 - n) / 2 + 2 * n
        self.assertEqual(n_terms, len(indices))
        self.assertEqual(n_terms, len(coefs))
        for i in range(len(indices)):
            idx = indices[i]
            if idx[0] == 0:
                term_iter = lin_terms[idx[1] - 1]
            elif idx[0] == idx[1]:
                term_iter = quad_terms[idx[0] - 1, idx[1] - 1]
            elif idx[0] != idx[1]:
                term_iter = quad_terms[idx[0] - 1, idx[1] - 1] * 2
            self.assertEqual(coefs[i], term_iter)

    def test_get_decimal_places(self):
        dec_5 = 1.23458
        dec_7 = 1.45678910
        dec_8 = 1.45678911
        dec_0 = 1
        dec_3_0 = 3.0
        self.assertEqual(get_decimal_places(dec_5), 5)
        self.assertEqual(get_decimal_places(dec_7), 7)
        self.assertEqual(get_decimal_places(dec_8), 8)
        self.assertEqual(get_decimal_places(dec_0), 0)
        self.assertEqual(get_decimal_places(dec_3_0), 0)



import sys

import numpy as np

from eqc_direct.client import EqcClient
from eqc_direct.utils import convert_hamiltonian_to_poly_format


def run_single(nvars):
    np.random.seed(13)
    ham_mat = np.abs(np.random.normal(size=(nvars, nvars + 1)))
    print("float 64:", ham_mat.flatten(order="F")[:10])
    poly_indices, poly_coefficients = convert_hamiltonian_to_poly_format(
        linear_terms=np.squeeze(ham_mat[:, 0]),
        quadratic_terms=ham_mat[:, 1:],
    )
    eqc_client = EqcClient()
    lock_id, _, _ = eqc_client.wait_for_lock()
    print("Lock:", lock_id)
    resp = eqc_client.solve_sum_constrained(
        poly_coefficients=poly_coefficients,
        poly_indices=poly_indices,
        sum_constraint=1,
        lock_id=lock_id,
    )
    print(resp)
    ham_mat_32 = ham_mat.astype(np.float32)
    print("float 32:", ham_mat.flatten(order="F")[0:10])
    poly_indices, poly_coefficients = convert_hamiltonian_to_poly_format(
        linear_terms=np.squeeze(ham_mat[:, 0]),
        quadratic_terms=ham_mat[:, 1:],
    )
    resp = eqc_client.solve_sum_constrained(
        poly_coefficients=poly_coefficients,
        poly_indices=poly_indices,
        sum_constraint=1,
        lock_id=lock_id,
    )
    print(resp)
    print("Float 32 type", ham_mat_32.dtype)
    eqc_client.release_lock(lock_id=lock_id)


if __name__ == "__main__":
    # Default num_variables is 5.
    nvars = int((sys.argv[1:2] or ["5"])[0])
    run_single(nvars=nvars)



"""
Simulates Server Wrapper for EQC3 (aka. Dirac-3) device
"""
from concurrent import futures
import logging
import math
import os
import signal
import threading
import time
from typing import Optional
import uuid

import grpc

# bypasses pickle serialization in favor of dill
# allows for consolidation of multiprocessing code in single class
# since multiprocessing blocks you from running multiprocessing code
# when module isn't main
import multiprocess
from multiprocess import managers
import numpy as np

import eqc_direct
from eqc_direct import eqc_pb2, eqc_pb2_grpc, utils

# max number of entries for matrix in c++ server
# based on 949 entries
MAX_ENTRIES = 949 * 949 + 949
MIN_CONSTRAINT = 1
MAX_CONSTRAINT = 10000


def calc_scale_factor(float_num: np.float32):
    decimal_places = len(str(np.float32(float_num)).split(".")[1])
    return 10**decimal_places


def convert_float_to_int(float_num: np.float32, scale_factor: int):
    int_number = int(float(str(float_num)) * scale_factor)
    return int_number


def calculate_mats_size(degree, num_variables):
    entries = 0
    for i in range(1, degree + 1):
        entries += num_variables**i
    return entries


def exceeds_max_size(degree: int, num_variables: int):
    # based on 949 variables for degree 2
    if degree <= 2:
        if num_variables > 949:
            return True
    elif degree == 3:
        if num_variables > 135:
            return True
    elif degree == 4:
        if num_variables > 39:
            return True
    elif degree == 5:
        if num_variables > 19:
            return True
    else:
        entries = calculate_mats_size(degree=degree, num_variables=num_variables)
        if entries > MAX_ENTRIES:
            return True
    return False


def check_for_duplicate_poly_index(poly_indices):
    seen = set()
    for sublist in poly_indices:
        if tuple(sublist) in seen:
            return True
        seen.add(tuple(sublist))
    return False


def check_decreasing(poly_indices):
    for sublist in poly_indices:
        for idx in range(len(sublist) - 1):
            if sublist[idx] > sublist[idx + 1]:
                return True
    return False


def calc_energy(poly_indices, poly_coefficients, soln):
    energy = 0
    for idx in range(len(poly_indices)):
        # start with initial coeffcient will be multiplied
        # by variable values for
        iter_energy = poly_coefficients[idx]
        iter_vars = poly_indices[idx]
        for var in iter_vars:
            if var > 0:
                iter_energy *= soln[(var - 1)]
        energy += iter_energy
    return energy


class EqcServer(eqc_pb2_grpc.EqcService):
    """
    Simulates an EQCX where X>=3
    Attributes
    ----------
    sys_tracker: dict with following keys
        - current_status: One of utils.SysStatus from utils.py
        - result: an EQC result object
        - idle_time: time from which EQC has been idle
        - current_pid: process id for job running on EQC
    spinlock: a lock for ensuring only single process can run at a time

    :note:
    context is the server stub.
    request is the input data for the grpc method used on client side.
    Server calibration occurs automatically when eqc-device is running
    and is not implemented in the server sim as there is no hardware.

    To see more info on request values see proto file
    """

    def __init__(self):
        # need to store as managed dict to access from other Processes
        self.manager = managers.SyncManager()
        self.manager.start()
        self.sys_tracker = self.manager.dict()
        self.sys_tracker.update(
            {
                "current_status": utils.SysStatus.IDLE,
                "result": {},
                "mixed_int_result": {},
                "idle_time": time.time(),
                "current_pid": None,
            }
        )
        self.spinlock = multiprocess.Lock()
        self.lock_id = ""
        # Unsure how this will performo windows so adding this as a check
        # If need a different start_method than default would use set_start_method
        logging.info("START TYPE %s", multiprocess.get_start_method())

    def __del__(self):
        self.manager.shutdown()

    def AcquireLock(self, request, context):
        """
        Releases spinlock if has been idle for more than 120 seconds
        :return: a bool indicating that spinlock was acquired or not
        """
        # In order to release lock if already locked must meet following:
        # 1. should have been idle for 120 seconds
        # 2. System must be currently idle
        # there are other checks for this last one so probably could be left out
        if (time.time() - self.sys_tracker["idle_time"]) > 120 and self.sys_tracker[
            "current_status"
        ] == utils.SysStatus.IDLE:
            self.sys_tracker["idle_time"] = time.time()
            try:
                self.spinlock.release()
            # if double release raises ValueError
            except ValueError:
                pass
        # race condition is avoided because only one can have spinlock
        lock_acquired = self.spinlock.acquire(block=False)
        if lock_acquired:
            lock_id = str(uuid.uuid4())
            self.lock_id = lock_id
            logging.info("Lock Acquired with lock_id: %s", self.lock_id)
            self.sys_tracker["idle_time"] = time.time()
            self.sys_tracker["result"] = {}
            lock_status = utils.LockManageStatus.SUCCESS
        else:
            lock_id = ""
            lock_status = utils.LockManageStatus.BUSY
        lock_out = {"lock_id": lock_id, "lock_status": lock_status}
        return eqc_pb2.LockOutput(**lock_out)

    def ReleaseLock(self, request, context):
        """
        Releases lock for execution
        """
        logging.info("Attempting lock release...")
        if request.lock_id != self.lock_id:
            logging.error("ReleaseLock failed because requested wrong lock_id")
            return eqc_pb2.StatusOutput(**utils.LockManageStatus.MISMATCH)
        if self.sys_tracker["current_pid"] != None:
            logging.error("ReleaseLock failed due to other process running")
            return eqc_pb2.StatusOutput(**utils.LockManageStatus.BUSY)
        try:
            # this may not be necessary
            self.lock_id = ""
            self.spinlock.release()
        # catch error for double release
        except RuntimeError:
            pass
        return eqc_pb2.StatusOutput(**utils.LockManageStatus.SUCCESS)

    def CheckLock(self, request, context):
        """
        Returns current lock availability to user
        """
        user_locked = self.lock_id == request.lock_id and self.lock_id != ""
        lock_available = self.lock_id == ""
        if user_locked:
            check_output = utils.LockCheckStatus.USER_LOCKED
        elif lock_available:
            check_output = utils.LockCheckStatus.AVAILABLE
        else:
            check_output = utils.LockCheckStatus.UNAVAILABLE
        return eqc_pb2.StatusOutput(**check_output)

    def SystemStatus(self, request, context):
        """
        Returns the current system status as a message
        """
        return eqc_pb2.StatusOutput(**self.sys_tracker["current_status"])

    def SystemInfo(self, request, context):
        """
        Returns server info with eqc-direct package version for Dirac-3 Python-based
        Dirac-3 software simulator.
        """

        return eqc_pb2.VersionOutput(
            server_version=eqc_direct.PACKAGE_VERSION,
            device_type="Dirac-3-simulator",
            fpga_version="None",
        )

    def set_idle(self):
        """
        Resets sys_tracker settings
        """
        self.sys_tracker["current_status"] = utils.SysStatus.IDLE
        self.sys_tracker["idle_time"] = time.time()
        self.sys_tracker["current_pid"] = None

    def SolveInt(self, request, context):
        """
        Runs simulated mixed-integer job
        """
        start_val = time.time()
        logging.info("Submit mixed-integer job...")
        job_input = utils.message_to_dict(request)
        # data validation
        if request.lock_id != self.lock_id:
            return eqc_pb2.SubmitOutput(**utils.JobCodes.LOCK_MISMATCH)
        # must not have another process running
        if self.sys_tracker["current_pid"] is not None:
            return eqc_pb2.SubmitOutput(**utils.JobCodes.DEVICE_BUSY)
        if (
            len(job_input["poly_indices"]) / len(job_input["coef_values"])
        ) != request.degree:
            return eqc_pb2.SubmitOutput(**utils.JobCodes.COEF_INDEX_MISMATCH)
        job_input["poly_indices"] = np.array(job_input["poly_indices"]).reshape(
            -1, job_input["degree"]
        )
        if request.num_variables < np.max(job_input["poly_indices"]):
            return eqc_pb2.SubmitOutput(**utils.JobCodes.INDEX_OUT_OF_RANGE)
        if job_input["num_samples"] <= 0:
            return eqc_pb2.SubmitOutput(**utils.JobCodes.NUM_SAMPLES_POSITIVE)
        if request.degree <= 0:
            return eqc_pb2.SubmitOutput(**utils.JobCodes.DEGREE_POSITIVE)
        if request.num_variables <= 0:
            return eqc_pb2.SubmitOutput(**utils.JobCodes.NUM_VARIABLES_POSITIVE)
        if job_input["relaxation_schedule"] not in {1, 2, 3, 4}:
            return eqc_pb2.SubmitOutput(**utils.JobCodes.INVALID_RELAXATION_SCHEDULE)
        if exceeds_max_size(degree=request.degree, num_variables=request.num_variables):
            return eqc_pb2.SubmitOutput(**utils.JobCodes.EXCEEDS_MAX_SIZE)
        if check_decreasing(job_input["poly_indices"]):
            return eqc_pb2.SubmitOutput(**utils.JobCodes.DECREASING_INDEX)
        if len(job_input["num_levels"]) != job_input["num_variables"]:
            return eqc_pb2.SubmitOutput(**utils.JobCodes.NUM_LEVELS_NUM_VARS_MISMATCH)
        if min(job_input["num_levels"]) <= 1:
            return eqc_pb2.SubmitOutput(**utils.JobCodes.NUM_LEVELS_GT_ONE)
        if sum(job_input["num_levels"]) > 954:
            return eqc_pb2.SubmitOutput(**utils.JobCodes.TOTAL_INTEGER_LEVELS)
        if job_input["mean_photon_number"] != 0 and (
            job_input["mean_photon_number"] > 0.0066667
            or job_input["mean_photon_number"] < 0.0000665
        ):
            return eqc_pb2.SubmitOutput(**utils.JobCodes.INVALID_MEAN_PHOTON_NUMBER)
        if job_input["quantum_fluctuation_coefficient"] != 0 and (
            job_input["quantum_fluctuation_coefficient"] > 100
            or job_input["quantum_fluctuation_coefficient"] < 1
        ):
            return eqc_pb2.SubmitOutput(
                **utils.JobCodes.INVALID_QUANTUM_FLUCTUATION_COEFFICIENT
            )
        self.sys_tracker["current_status"] = utils.SysStatus.JOB_RUNNING
        # Python deserialize it as type float which defaults to 64 bit
        # precision in python in order to ensure that it has the same
        # precision as the C++ server use np.float32
        job_input["coef_values"] = np.array(job_input["coef_values"], dtype=np.float32)
        end_val = time.time()
        job_input["preprocessing_time"] = end_val - start_val
        logging.info("Start subprocess...")
        proc_job = multiprocess.Process(
            target=self.run_mixed_integer, args=(job_input,)
        )
        proc_job.start()
        self.sys_tracker["current_pid"] = proc_job.pid
        return eqc_pb2.SubmitOutput(**utils.JobCodes.NORMAL)

    def run_mixed_integer(self, job_input):
        """
        Simulates mixed-integer job
        """
        result = {
            "preprocessing_time": job_input["preprocessing_time"],
            "runtime": [],
            "postprocessing_time": [],
            "energy": [],
            "solution": [],
            "num_samples": job_input["num_samples"],
            "num_variables": job_input["num_variables"],
            "calibration_time": 0,
        }

        logging.info("Run job...")
        for _ in range(job_input["num_samples"]):
            run_start = time.time()
            sim_solution = np.array(
                [
                    # Currently, we simulate only integer solutions.
                    np.random.randint(int(math.floor(num_levels)))
                    for num_levels in job_input["num_levels"]
                ],
                dtype=np.float32,  # Matches type and precision from Dirac.
            )
            post_start = time.time()
            energy = calc_energy(
                poly_indices=job_input["poly_indices"],
                poly_coefficients=job_input["coef_values"],
                soln=sim_solution,
            )
            post_end = time.time()
            result["runtime"].append(post_start - run_start)
            result["postprocessing_time"].append(post_end - post_start)
            result["energy"].append(energy)
            # wait to convert to list since distill process requires numpy array
            result["solution"].append({"values": sim_solution.tolist()})

        result.update(utils.JobCodes.NORMAL)
        self.sys_tracker["mixed_int_result"] = result
        self.set_idle()
        return result

    def SolveSumConst(self, request, context):
        """
        Runs sum-constrained job simulation
        """
        start_val = time.time()
        logging.info("Submit sum-constrained job...")
        job_input = utils.message_to_dict(request)
        # data validation
        if request.lock_id != self.lock_id:
            return eqc_pb2.SubmitOutput(**utils.JobCodes.LOCK_MISMATCH)
        # must not have another process running
        if self.sys_tracker["current_pid"] is not None:
            return eqc_pb2.SubmitOutput(**utils.JobCodes.DEVICE_BUSY)
        # check sum constraint
        scale_factor = max(
            calc_scale_factor(np.float32(request.soln_precision)),
            calc_scale_factor(np.float32(request.sum_constraint)),
        )
        prec_int = convert_float_to_int(
            np.float32(request.soln_precision), scale_factor
        )
        constraint_int = convert_float_to_int(
            np.float32(request.sum_constraint), scale_factor
        )
        if (
            len(job_input["poly_indices"]) / len(job_input["coef_values"])
        ) != request.degree:
            return eqc_pb2.SubmitOutput(**utils.JobCodes.COEF_INDEX_MISMATCH)
        job_input["poly_indices"] = np.array(job_input["poly_indices"]).reshape(
            -1, job_input["degree"]
        )
        if request.num_variables < np.max(job_input["poly_indices"]):
            return eqc_pb2.SubmitOutput(**utils.JobCodes.INDEX_OUT_OF_RANGE)
        if job_input["num_samples"] <= 0:
            return eqc_pb2.SubmitOutput(**utils.JobCodes.NUM_SAMPLES_POSITIVE)
        if request.soln_precision < 0:
            return eqc_pb2.SubmitOutput(**utils.JobCodes.PRECISION_NONNEGATIVE)
        if request.degree <= 0:
            return eqc_pb2.SubmitOutput(**utils.JobCodes.DEGREE_POSITIVE)
        if request.num_variables <= 0:
            return eqc_pb2.SubmitOutput(**utils.JobCodes.NUM_VARIABLES_POSITIVE)
        if request.sum_constraint < 1 or request.sum_constraint > 10000:
            return eqc_pb2.SubmitOutput(**utils.JobCodes.INVALID_SUM_CONSTRAINT)
        if request.soln_precision != 0:
            if (constraint_int / prec_int) > 10000:
                return eqc_pb2.SubmitOutput(**utils.JobCodes.INVALID_PRECISION)
            if not ((constraint_int % prec_int) == 0):
                return eqc_pb2.SubmitOutput(
                    **utils.JobCodes.PRECISION_CONSTRAINT_MISMATCH
                )
        if job_input["relaxation_schedule"] not in {1, 2, 3, 4}:
            return eqc_pb2.SubmitOutput(**utils.JobCodes.INVALID_RELAXATION_SCHEDULE)
        if exceeds_max_size(degree=request.degree, num_variables=request.num_variables):
            return eqc_pb2.SubmitOutput(**utils.JobCodes.EXCEEDS_MAX_SIZE)
        if check_decreasing(job_input["poly_indices"]):
            return eqc_pb2.SubmitOutput(**utils.JobCodes.DECREASING_INDEX)
        print("MPN:", job_input["mean_photon_number"])
        if job_input["mean_photon_number"] != 0 and (
            job_input["mean_photon_number"] > 0.0066667
            or job_input["mean_photon_number"] < 0.0000665
        ):
            return eqc_pb2.SubmitOutput(**utils.JobCodes.INVALID_MEAN_PHOTON_NUMBER)
        if job_input["quantum_fluctuation_coefficient"] != 0 and (
            job_input["quantum_fluctuation_coefficient"] > 100
            or job_input["quantum_fluctuation_coefficient"] < 1
        ):
            return eqc_pb2.SubmitOutput(
                **utils.JobCodes.INVALID_QUANTUM_FLUCTUATION_COEFFICIENT
            )
        self.sys_tracker["current_status"] = utils.SysStatus.JOB_RUNNING
        # Python deserialize it as type float which defaults to 64 bit
        # precision in python in order to ensure that it has the same
        # precision as the C++ server use np.float32
        job_input["coef_values"] = np.array(job_input["coef_values"], dtype=np.float32)
        end_val = time.time()
        job_input["preprocessing_time"] = end_val - start_val
        logging.info("Start subprocess...")
        proc_job = multiprocess.Process(target=self.run_sum_const, args=(job_input,))
        proc_job.start()
        self.sys_tracker["current_pid"] = proc_job.pid
        return eqc_pb2.SubmitOutput(**utils.JobCodes.NORMAL)

    def run_sum_const(self, job_input):
        """
        Simulates running sum-constrained job
        :note: doesn't simulate distillation process so not exact simulation
        """
        result = {
            "preprocessing_time": job_input["preprocessing_time"],
            "runtime": [],
            "postprocessing_time": [],
            "energy": [],
            "solution": [],
            "distilled_energy": [],
            "distilled_solution": [],
            "num_samples": job_input["num_samples"],
            "num_variables": job_input["num_variables"],
            "calibration_time": 0,
        }

        logging.info("Run sum-constrained job...")
        for _ in range(job_input["num_samples"]):
            start_time = time.time()
            sim_solution = np.random.randint(10000, size=(job_input["num_variables"],))
            sim_solution = (
                job_input["sum_constraint"] * sim_solution / np.sum(sim_solution)
            ).astype(
                np.float32
            )  # Matches type and precision from Dirac.
            energy = calc_energy(
                poly_indices=job_input["poly_indices"],
                poly_coefficients=job_input["coef_values"],
                soln=sim_solution,
            )
            end_runtime = time.time()
            runtime = end_runtime - start_time
            result["runtime"].append(runtime)
            result["energy"].append(energy)
            if job_input["soln_precision"] != 0:
                rounded_solution = (
                    np.floor(sim_solution / job_input["soln_precision"])
                    * job_input["soln_precision"]
                )
                p = np.round(
                    (job_input["sum_constraint"] - rounded_solution.sum())
                    / job_input["soln_precision"]
                )
                rounded_solution[0] += p * job_input["soln_precision"]
                dist_energy = calc_energy(
                    poly_indices=job_input["poly_indices"],
                    poly_coefficients=job_input["coef_values"],
                    soln=rounded_solution,
                )
                dist_solution = [np.float32(var) for var in rounded_solution.tolist()]
                result["distilled_solution"].append({"values": dist_solution})
                result["distilled_energy"].append(dist_energy)
            # wait to convert to list since distill process requires numpy array
            sim_solution = [np.float32(var) for var in sim_solution.tolist()]

            result["solution"].append({"values": sim_solution})
            result["postprocessing_time"].append(time.time() - end_runtime)

        result.update(utils.JobCodes.NORMAL)
        self.sys_tracker["result"] = result
        self.set_idle()
        return result

    def FetchSumConstResults(self, request, context):
        """
        Fetch most recent sum-constrained results
        """
        if (
            self.lock_id == request.lock_id
            and self.sys_tracker["current_status"] == utils.SysStatus.IDLE
        ):
            logging.info("Fetch results...")
            return eqc_pb2.SumConstResult(**self.sys_tracker["result"])
        elif self.lock_id != request.lock_id:
            return eqc_pb2.SumConstResult(**utils.JobCodes.LOCK_MISMATCH)
        else:
            return eqc_pb2.SumConstResult(**utils.JobCodes.DEVICE_BUSY)

    def FetchIntResults(self, request, context):
        """
        Fetch most recent mixed-integer result
        """
        if (
            self.lock_id == request.lock_id
            and self.sys_tracker["current_status"] == utils.SysStatus.IDLE
        ):
            logging.info("Fetch mixed-integer results...")
            return eqc_pb2.IntResult(**self.sys_tracker["mixed_int_result"])
        elif self.lock_id != request.lock_id:
            return eqc_pb2.IntResult(**utils.JobCodes.LOCK_MISMATCH)
        else:
            return eqc_pb2.IntResult(**utils.JobCodes.DEVICE_BUSY)

    def StopRunning(self, request, context):
        """
        Stops a running job when using simulator may not work against actual server code
        """
        logging.info("Stop current running process...")
        if request.lock_id != self.lock_id:
            return eqc_pb2.StatusOutput(**utils.LockManageStatus.MISMATCH)
        if self.sys_tracker["current_pid"] is not None:
            logging.info("Killing process: %s", self.sys_tracker["current_pid"])
            os.kill(self.sys_tracker["current_pid"], signal.SIGKILL)
            self.set_idle()
            self.sys_tracker["result"] = utils.JobCodes.USER_INTERRUPT
        return eqc_pb2.StatusOutput(**utils.LockManageStatus.SUCCESS)


class GrpcServer:
    """
    Provides basic functionality to start a threaded gRPC server instance
    this is used to run the server simulator

    :param ip_address: target ip address to start grpc server
    :param port: target port on which to start grpc server
    :param private_key: a private key pem file that is used to run TLS
    :param cert_file: a certificate pem file that is used to run TLS
    """

    def __init__(
        self,
        ip_address: str = "localhost",
        port: str = "50051",
        private_key: Optional[str] = None,
        cert_file: Optional[str] = None,
    ):
        os.environ["GRPC_ENABLE_FORK_SUPPORT"] = "true"
        os.environ["GRPC_POLL_STRATEGY"] = "poll"
        self.stop_event = threading.Event()
        server_cls = EqcServer()
        max_variables = 10000
        max_length = (
            int(
                ((((4 * max_variables**2)) + (4 * max_variables)) / (1024**2))
                + (
                    512
                    - (
                        (((4 * max_variables**2) + (4 * max_variables)) / (1024**2))
                        % 512
                    )
                )
            )
            * 1024**2
        )
        self.server = grpc.server(
            futures.ThreadPoolExecutor(max_workers=10),
            options=(
                ("grpc.max_receive_message_length", max_length),
                ("grpc.max_send_message_length", max_length),
            ),
        )

        eqc_pb2_grpc.add_EqcServiceServicer_to_server(server_cls, self.server)
        ip_address = ip_address if ip_address else "localhost"
        port = port if port else "50051"
        logging.info("grpc server ip_address:port %s:%s", ip_address, port)
        ip_add_port = f"{ip_address}:{port}"
        if private_key and cert_file:
            with open(cert_file, "rb") as f:
                server_cert = f.read()
            with open(private_key, "rb") as f:
                server_key = f.read()
            server_credentials = grpc.ssl_server_credentials(
                [(server_key, server_cert)]
            )
            self.server.add_secure_port(ip_add_port, server_credentials)
        else:
            self.server.add_insecure_port(ip_add_port)

    def serve(self):
        """
        Starts grpc server instance which waits for termination
        """
        logging.info("Server started")
        self.server.start()
        self.server.wait_for_termination()

    def stop(self):
        """
        Stops the classes running grpc server instance
        with a stop event
        """
        self.server.stop(0)
        print("Stopping server")
        self.stop_event.set()



import sys
import time
from multiprocessing import Pool
from eqc_direct.eqc_client import EqcClient
import numpy as np


def run_single():
    np.random.seed(13)
    ham_mat = np.load("enlight8_alpha1.0.npy")
    print(ham_mat)
    print(ham_mat.shape)
    start = time.time()
    eqc_client = EqcClient()
    lock_id, start_ts, end_ts = eqc_client.wait_for_lock()
    print("Lock:", lock_id)
    resp = eqc_client.process_job(
        hamiltonian=ham_mat,
        relaxation_schedule=2,
        sum_constraint=30,
        continuous_soln=False,
        lock_id=lock_id,
    )
    eqc_client.release_lock(lock_id=lock_id)
    print(resp)


if __name__ == "__main__":
    run_single()



"""
:class:`.EqcClient` contains all gRPC calls to solve problems and monitor system
"""
import json
import logging
import time
from typing import List, Optional, TypedDict, Union
import warnings

import grpc
from grpc._channel import _InactiveRpcError
import numpy as np
import numpy.typing as np_typing

import eqc_direct
from eqc_direct import eqc_pb2, eqc_pb2_grpc
from eqc_direct.utils import SysStatus, SystemInfo, get_decimal_places, message_to_dict

DIRAC_FLOAT_BITS = 32


class InactiveRpcError(Exception):
    """Custom exception wrapper around grpc._channel._InactiveRpcError."""


class SumConstrainedResult(TypedDict):
    """
    Sum-constrained results object. Will not contain energy or solution if err_code is
    not 0. Contains nonempty distilled entries when solution_precision is not 0 or None.

    :param err_code: the error code for a given job. Full list of :code:`err_code`
        values can be found :class:`eqc_direct.utils.JobCodes`
    :param err_desc: the error description for a given job submission. Full list of
        :code:`err_desc` values can be found  in :class:`eqc_direct.utils.JobCodes`
    :param preprocessing_time: data validation and time to re-format input data for
        running on the device in seconds
    :param runtime: sampling time in seconds for each solve on Dirac hardware
    :param energy: list of energies for best solution found (float32 precision) for
        each sample from Dirac hardware
    :param solution: a list of vectors representing the lowest energy solution
        (float32 precision) for each sample from Dirac hardware
    :param postprocessing_time: runtime for auxilary computations that occur
        besides sampling during each sample routine from Dirac hardware including
        intermediate energy calculations, objective function adjustments, and
        distillation of solutions
    :param distilled_energy: list of energies for distilled solution for input polynomial
        (float32 precision) for each sample from Dirac hardware
    :param distilled_solution: a vector representing the solution after
        the distillation procedure is applied to the original solution
        derived from the hardware. (float32 precision)
    :param calibration_time: calibration time is unrelated to execution of the
        individual sampling for the optimization. This time is from system level
        interruptions from calibrations that happen at regular intervals to maintain
        system performance.

    .. note::
      * solutions are length `num_variables` vectors of non-negative float that sum to
        `sum_constraint`
      * distilled solutions, when present, are length `num_variables` vectors of
        non-negative float with precision `solution_precision` that sum to
        `sum_constraint`
    """

    err_code: int
    err_desc: str
    num_samples: int
    num_variables: int
    preprocessing_time: float
    runtime: List[float]
    postprocessing_time: List[float]
    energy: List[float]
    solution: List[List[float]]
    distilled_energy: List[float]
    distilled_solution: List[List[float]]
    calibration_time: float


class IntegerResult(TypedDict):
    """
    Integer solver results object. Will not contain energy or solution if err_code is not
    0.

    :param err_code: the error code for a given job. Full list of :code:`err_code`
        values can be found :class:`eqc_direct.utils.JobCodes`
    :param err_desc: the error description for a given job submission. Full list of
        :code:`err_desc` values can be found  in :class:`eqc_direct.utils.JobCodes`
    :param preprocessing_time: data validation and time to re-format input data for
        running on the device in seconds
    :param runtime: sampling time in seconds for each solve on Dirac hardware
    :param postprocessing_time: runtime for auxilary computations that occur
        besides sampling during each sample routine from Dirac hardware including
        intermediate energy calculations and objective function adjustments
    :param energy: list of energies of best solution found for each sample from Dirac hardware
    :param solution: list of vectors representing the lowest energy solution found by the device
        for each sample from Dirac hardware
    :param calibration_time: calibration time is unrelated to execution of the individual
        sampling for the optimization but it will occassionally interrupt execution to
        run calibration to maintain the system in order to provide transparency for total
        where time was spent solving a specific problem this value is provided


    .. note::
      * solutions are length `num_variables` vectors of non-negative bounded above
        inclusively by `num_levels`-1.
    """

    err_code: int
    err_desc: str
    num_samples: int
    num_variables: int
    preprocessing_time: List[float]
    runtime: List[float]
    postprocessing_time: List[float]
    energy: List[float]
    solution: List[List[float]]


class EqcClient:
    """
    Client interface to communicate with EQC device.

    :param ip_address: The IP address of the EQC device
    :param port: The port over which the gRPC server is communicating on the EQC device

    .. note::

       :code:`lock_id` is used by a variety of class functions.
       It is set to an empty string by default since default for device server
       :code:`lock_id` is also an empty string. This allows for single user
       processing without having to acquire a device lock.

    .. .. admonition::
       All GRPC calls follow a specific pattern:
       1. Fill in data to be sent in message stub
       2. Send data using stub service method
       3. Parse response
    """

    def __init__(
        self,
        ip_address: str = "",
        port: str = "",
        cert_file: Optional[str] = None,
    ):
        if not ip_address:
            ip_address = eqc_direct.DEVICE_IP_ADDRESS_DEFAULT

        self._ip_address = ip_address

        if not port:
            port = eqc_direct.DEVICE_PORT_DEFAULT

        self._port = port

        max_data_size = 512 * 1024 * 1024
        options = [
            ("grpc.max_send_message_length", max_data_size),
            ("grpc.max_receive_message_length", max_data_size),
            (
                "grpc.service_config",
                json.dumps(
                    {
                        "methodConfig": [
                            {
                                "name": [{"service": "eqc.EqcService"}],
                                "retryPolicy": {
                                    "maxAttempts": 5,
                                    "initialBackoff": "0.2s",
                                    "maxBackoff": "10s",
                                    "backoffMultiplier": 2.5,
                                    "retryableStatusCodes": ["UNAVAILABLE"],
                                },
                            }
                        ]
                    }
                ),
            ),
        ]

        if cert_file:
            # read in certs
            with open(cert_file, "rb") as f:
                ca_cert = f.read()

            channel_credentials = grpc.ssl_channel_credentials(ca_cert)
            self.channel = grpc.secure_channel(
                self.ip_add_port, channel_credentials, options=options
            )
        else:
            self.channel = grpc.insecure_channel(self.ip_add_port, options=options)

        self.eqc_stub = eqc_pb2_grpc.EqcServiceStub(self.channel)

    @property
    def ip_address(self) -> str:
        """Return configured IP address."""
        return self._ip_address

    @property
    def port(self) -> str:
        """Return configured IP port."""
        return self._port

    @property
    def ip_add_port(self) -> str:
        """Return configured IP address + port, concatenated with a colon."""
        return self.ip_address + ":" + self.port

    def submit_integer_job(
        self,
        poly_coefficients: Union[List[float], np_typing.NDArray[np.float32]],
        poly_indices: Union[List[List[int]], np_typing.NDArray[np.uint32]],
        num_levels: Union[int, List[int], np_typing.NDArray[np.uint16]],
        num_variables: Optional[int] = None,
        num_samples: int = 1,
        lock_id: str = "",
        relaxation_schedule: int = 1,
        mean_photon_number: Optional[float] = None,
        quantum_fluctuation_coefficient: Optional[int] = None,
    ) -> dict:
        """
        Submits a polynomial to be minimized by Dirac, with solutions restricted to integers
        between 0 and :math:`L - 1`, where :math:`L` is the value in :code:`num_levels` for each
        variable.

        :param poly_coefficients:
            Coefficient values for the polynomial to be minimized.
            Inputs should be precision float32 or less, as higher precision will be truncated.
        :param poly_indices:
            Index sets corresponding to the variables for each coefficient in the polynomial.
        :param num_levels:
            Specifies the number of discrete values (levels) each variable can take.
            A value of 2 means a variable can be {0, 1}, and in general, if :math:`L_i` is
            specified for variable :math:`i`, then its domain is :math:`\{0, 1, ..., L_i - 1\}`.

            .. math::
                \\sum_{i=1}^{n} L_i \\leq 954

        :param num_variables:
            Optional, specifies the number of variables in the polynomial. Must be greater than or
            equal to the maximum index appearing in :code:`poly_indices`.
        :param num_samples:
            The number of samples (independent solutions) to generate from the device.
        :param lock_id:
            A UUID to coordinate multi-user access to the device.
        :param relaxation_schedule:
            An integer from the set {1, 2, 3, 4} indicating the type of analog relaxation schedule.
            Higher values reduce the variability in analog spin values, increasing the probability
            of reaching lower-energy solutions.
        :param mean_photon_number:
            Optional, overrides the default photon number associated with the selected
            :code:`relaxation_schedule`. This sets the average number of photons present in a
            given quantum state.
        :param quantum_fluctuation_coefficient:
            Optional, overrides the default value from :code:`relaxation_schedule`. Specifies the
            number of photons (:math:`N`) in each feedback loop, which determines
            the shot noise. Accepts integer values in the range [0, 100].
            Shot noise scales inversely with the square root of the photon number:

        .. math::
            \\delta x \\propto \\frac{1}{\\sqrt{N}}

        :return:
            A dictionary from :class:`eqc_direct.utils.JobCodes` with the following keys:

            - **err_code** (`int`): Error code for the job submission.
            - **err_desc** (`str`): Description corresponding to the error code.
        """
        poly_coefficients = np.array(poly_coefficients)
        poly_coefficients_dtype = poly_coefficients.dtype

        if not (
            np.issubdtype(poly_coefficients_dtype, np.floating)
            and np.finfo(poly_coefficients_dtype).bits <= DIRAC_FLOAT_BITS
        ):
            warn_dtype_msg = (
                "Max precision for `poly_coefficients` representation in EQC device is "
                f"float32, but input dtype was {poly_coefficients_dtype.name}. Input "
                "polynomial coefficients will be rounded."
            )
            logging.warning(warn_dtype_msg)
            warnings.warn(warn_dtype_msg, Warning)

        poly_indices = np.array(poly_indices)

        try:
            _, poly_degree = poly_indices.shape
        except ValueError as err:
            err_msg = "`poly_indices` array must be two dimensional"
            logging.error(err_msg, exc_info=True)
            raise ValueError(err_msg) from err

        if num_variables is None:
            num_variables = np.max(poly_indices)

        if mean_photon_number is None:
            mean_photon_number = 0

        if quantum_fluctuation_coefficient is None:
            quantum_fluctuation_coefficient = 0

        # Scalar num_levels indicates to use same num_levels for each variable.
        # This "passes through" if num_levels as an array.
        num_levels = np.broadcast_to(num_levels, (num_variables,))

        if not (np.issubdtype(num_levels.dtype, np.integer)):
            raise ValueError("`num_levels` must be type int")

        logging.info("Submitting integer job to device...")
        job_input = eqc_pb2.IntInput(
            num_variables=num_variables,
            degree=poly_degree,
            poly_indices=poly_indices.flatten(order="c").tolist(),  # rowwise for matrix
            coef_values=poly_coefficients.tolist(),
            num_levels=num_levels.tolist(),
            mean_photon_number=mean_photon_number,
            quantum_fluctuation_coefficient=quantum_fluctuation_coefficient,
            relaxation_schedule=relaxation_schedule,
            num_samples=num_samples,
            lock_id=lock_id,
        )
        logging.info("Submitting integer job to device...done")

        try:
            job_results = self.eqc_stub.SolveInt(job_input)
        except _InactiveRpcError as exc:
            # Make error easier to read/detect by consumers.
            raise InactiveRpcError(
                "EQC submit_job failed due to grpc._channel._InactiveRpcError."
            ) from exc

        return message_to_dict(job_results)

    def fetch_integer_result(self, lock_id: str = "") -> IntegerResult:
        """
        Fetches results for last integer solve on Dirac device.

        :param lock_id: a valid :code:`lock_id` that matches current device
            :code:`lock_id`
        :return: an :class:`.IntegerResult` object
        """
        try:
            logging.info("Fetching integer results...")
            integer_results = self.eqc_stub.FetchIntResults(
                eqc_pb2.LockMessage(lock_id=lock_id)
            )
            logging.info("Fetching integer results...done")
        except _InactiveRpcError as exc:
            # Make error easier to read/detect by consumers.
            raise InactiveRpcError(
                "EQC fetch_results failed due to grpc._channel._InactiveRpcError."
            ) from exc

        result = message_to_dict(integer_results)

        if result["solution"]:
            result["solution"] = np.array(result["solution"])
            result["solution"] = [
                [float(f"{np.float32(val):.7f}") for val in soln]
                for soln in result["solution"]
            ]

            result["energy"] = [
                float(f"{np.float32(energy):.7f}") for energy in result["energy"]
            ]

        return result

    def solve_integer(
        self,
        poly_coefficients: Union[List[float], np_typing.NDArray[np.float32]],
        poly_indices: Union[List[List[int]], np_typing.NDArray[np.uint32]],
        num_levels: Union[int, List[int], np_typing.NDArray[np.uint16]],
        num_variables: Optional[int] = None,
        num_samples: int = 1,
        lock_id: str = "",
        relaxation_schedule: int = 1,
        mean_photon_number: Optional[float] = None,
        quantum_fluctuation_coefficient: Optional[int] = None,
    ) -> dict:
        """
        Utilizes Dirac to optimize a polynomial with integer solution values.

        :param poly_coefficients:
            the coefficient values for polynomial to be minimized. Inputs
            should be precision float 32 or less, otherwise precision is
            lost during conversion to 32-bit.
        :param poly_indices:
            the indices for coefficient values for polynomial to be minimized.
        :param num_levels: an array indicating the number of integer values
            for each solution variable. A :code:`num_levels` value of 2 for a variable
            indicates that the possible values for that variable can be {0,1}.
            Similarily if :math`L` is specified for a given variable then possible values
            for that variable will be {0,1, ..., :math:`L`-1}.

            .. math::
               \sum_{i=1}^{n} L_i \le 954

        :param num_variables: optional input to specify number of variables for
            polynomial. Must be greater than or equal to maximum index value in
            :code:`poly_indices`.
        :param num_samples: the number of times to solve the problem on the device.
        :param lock_id: a UUID to control multi-user device access.
        :param relaxation_schedule: four different schedules represented
            in integer parameter. Higher values reduce the variation in
            the analog spin values and therefore, are more probable to lead to
            improved (i.e., lower) objective function energy for input problem.
            Accepts range of values in set {1, 2, 3, 4}.
        :param mean_photon_number: optional parameter that modfies device
            configuration from the defaults for :code:`relaxation_schedule`.
            Sets the average number of photons that are present in a given
            quantum state.
        :param quantum_fluctuation_coefficient:
            Optional, overrides the default value from :code:`relaxation_schedule`. Specifies the
            number of photons (:math:`N`) in each feedback loop, which determines
            the shot noise. Accepts integer values in the range [0, 100].
            Shot noise scales inversely with the square root of the photon number:

        .. math::
            \\delta x \\propto \\frac{1}{\\sqrt{N}}

        :return: dict of results and timings with all keys from
            :class:`.IntegerResult` as well as the following additional keys:

           - start_job_ts: time in ns marking start of job_submission
           - end_job_ts: time in ns marking end of job submission complete includes

        :note: The difference between `end_job_ts` and `start_job_ts` includes a 1
           second polling time which is not part of device solving time. To calculate
           the execution time on the device add together `preprocessing_time`, `runtime`
           and `postprocessing_time` from results object samplewise.
        """
        start_job_ts = time.time_ns()
        submit_job_resp = self.submit_integer_job(
            poly_coefficients=poly_coefficients,
            poly_indices=poly_indices,
            num_levels=num_levels,
            num_variables=num_variables,
            num_samples=num_samples,
            lock_id=lock_id,
            relaxation_schedule=relaxation_schedule,
            mean_photon_number=mean_photon_number,
            quantum_fluctuation_coefficient=quantum_fluctuation_coefficient,
        )
        if submit_job_resp["err_code"] != 0:
            err_msg = f"Submission failed with response: {submit_job_resp}"
            logging.error(err_msg, exc_info=True)
            raise RuntimeError(err_msg)
        sys_code = self.system_status()["status_code"]
        while sys_code != SysStatus.IDLE["status_code"]:
            sys_code = self.system_status()["status_code"]
            # this is based on the error statuses are 3 and above
            if sys_code > 3:
                err_msg = f"System unavailable status_code: {sys_code}"
                logging.error(err_msg, exc_info=True)
                raise RuntimeError(err_msg)
            # only sleep if not idle
            if sys_code != SysStatus.IDLE["status_code"]:
                time.sleep(1)
        # pull in results after is idle
        job_result = self.fetch_integer_result(lock_id=lock_id)
        end_job_ts = time.time_ns()
        if job_result["err_code"] != 0:
            raise RuntimeError(
                "Job execution error\n"
                f"err_code: {job_result['err_code']}\n"
                f"err_desc: {job_result['err_desc']}"
            )
        job_result["start_job_ts"] = start_job_ts
        job_result["end_job_ts"] = end_job_ts

        return job_result

    def submit_sum_constrained_job(  # pylint: disable=R0913, R0914
        self,
        poly_coefficients: Union[List[float], np_typing.NDArray[np.float32]],
        poly_indices: Union[List[List[int]], np_typing.NDArray[np.uint32]],
        num_variables: Optional[int] = None,
        num_samples: int = 1,
        lock_id: str = "",
        relaxation_schedule: int = 1,
        sum_constraint: Union[int, float] = 10000,
        solution_precision: Optional[float] = None,
        mean_photon_number: Optional[float] = None,
        quantum_fluctuation_coefficient: Optional[int] = None,
    ) -> dict:
        """
        Submits polynomial to be minimized by Dirac. All solutions are optimized
        utilizing a sum constraint which limits search to only solutions which have a
        sum equal to the input sum constraint.

        :param poly_coefficients:
            the coefficient values for polynomial to be minimized. Numbers, including
            integers, should be floats with 32-bit (or less) precision, otherwise
            precision is lost during conversion to 32-bit.
        :param poly_indices:
            the indices for coefficient values for polynomial to be minimized.
        :param num_variables: optional input to specify number of variables for
            polynomial. Must be greater than or equal to maximum index value in
            :code:`poly_indices`.
        :param num_samples: the number of times to solve the problem on the device.
        :param lock_id: a UUID to allow for multi-user processing
        :param relaxation_schedule: four different schedules represented
            in integer parameter. Higher values reduce the variation in
            the analog spin values and therefore, are more probable to lead to
            improved objective function energy for input problem.
            Accepts range of values in set {1, 2, 3, 4}.
        :param sum_constraint: a normalization constraint that is applied to the
            problem space that is used to calculate :code:`energy`. This
            parameter will be rounded if exceeds float32 precision
            (e.g. 7-decimal places). Value must be between 1 and 10000.
        :param solution_precision: the level of precision to apply to the solutions.
            This parameter will be rounded if exceeds float32 precision
            (e.g. 7-decimal places). If specified a distillation method is
            applied to the continuous solutions to map them to the submitted
            :code:`solution_precision`. Input :code:`solution_precision` must
            satisfy :code:`solution_precision` greater than or equal to
            :code:`sum_constraint`/10000 in order to be valid.
            Also :code:`sum_constraint` must be divisible by :code:`solution_precision`.
            If :code:`solution_precision` is not specified no distillation will be
            applied to the solution derived by the device.
        :param mean_photon_number: optional parameter that modfies device
            configuration from the defaults for :code:`relaxation_schedule`.
            Sets the average number of photons that are present in a given
            quantum state.
        :param quantum_fluctuation_coefficient:
            Optional, overrides the default value from :code:`relaxation_schedule`. Specifies the
            number of photons (:math:`N`) in each feedback loop, which determines
            the shot noise. Accepts integer values in the range [0, 100].
            Shot noise scales inversely with the square root of the photon number:

        .. math::
            \\delta x \\propto \\frac{1}{\\sqrt{N}}

        :return: a member of :class:`eqc_direct.utils.JobCodes` as a dict
           with the following keys:

           - **err_code**: `int`- job submission error code
           - **err_desc**: `str`- error code description for submission
        """
        # set None values to zero for grpc messages
        if solution_precision is None:
            solution_precision = 0

        if mean_photon_number is None:
            mean_photon_number = 0

        if quantum_fluctuation_coefficient is None:
            quantum_fluctuation_coefficient = 0

        poly_coefficients = np.array(poly_coefficients)
        poly_indices = np.array(poly_indices)
        coefficient_dtype = poly_coefficients.dtype

        if not (
            np.issubdtype(coefficient_dtype, np.integer)
            or (
                np.issubdtype(coefficient_dtype, np.floating)
                and np.finfo(coefficient_dtype).bits <= DIRAC_FLOAT_BITS
            )
        ):
            warn_dtype_msg = (
                "Max precision for EQC device is float32 input type "
                f"was dtype {np.dtype(coefficient_dtype).name}."
                " Input matrix will be rounded"
            )
            logging.warning(warn_dtype_msg)
            warnings.warn(warn_dtype_msg, Warning)

        if get_decimal_places(solution_precision) > 7:
            soln_prec_warn = (
                "`solution_precision`precision is greater than 7 "
                "decimal places. Will be modified on submission to "
                "device to float32 precision"
            )
            logging.warning(soln_prec_warn)
            warnings.warn(soln_prec_warn, Warning)

        if get_decimal_places(sum_constraint) > 7:
            sum_constraint_warn = (
                "`sum_constraint` precision is greater than 7 decimal "
                "places. Will be modified on submission to device "
                "to float32"
            )
            logging.warning(sum_constraint_warn)
            warnings.warn(sum_constraint_warn, Warning)

        try:
            _, degree_poly = poly_indices.shape
        except ValueError as err:
            err_msg = "`poly_indices` array must be two dimensions"
            logging.error(err_msg, exc_info=True)
            raise ValueError(err_msg) from err

        if not num_variables:
            num_variables = np.max(poly_indices)

        logging.info("Submitting sum-constrained job to device...")
        job_input = eqc_pb2.SumConstInput(
            num_variables=num_variables,
            num_samples=num_samples,
            degree=degree_poly,
            poly_indices=poly_indices.flatten(
                order="c"
            ).tolist(),  # flatten rowwise for matrix
            coef_values=poly_coefficients.tolist(),
            sum_constraint=sum_constraint,
            relaxation_schedule=relaxation_schedule,
            soln_precision=solution_precision,
            mean_photon_number=mean_photon_number,
            quantum_fluctuation_coefficient=quantum_fluctuation_coefficient,
            lock_id=lock_id,
        )
        logging.info("Submitting sum-constrained job to device...done")

        try:
            job_results = self.eqc_stub.SolveSumConst(job_input)
        except _InactiveRpcError as exc:
            # Make error easier to read/detect by consumers.
            raise InactiveRpcError(
                "EQC submit_job failed due to grpc._channel._InactiveRpcError."
            ) from exc
        return message_to_dict(job_results)

    def fetch_sum_constrained_result(self, lock_id: str = "") -> SumConstrainedResult:
        """
        Request last EQC job results. Returns results from the most recent
        run on the device.

        :param lock_id: a valid :code:`lock_id` that matches current device
            :code:`lock_id`
        :return: an :class:`.EqcResult` object
        """
        fetch_input = eqc_pb2.LockMessage(lock_id=lock_id)
        try:
            eqc_results = self.eqc_stub.FetchSumConstResults(fetch_input)
        except _InactiveRpcError as exc:
            # Make error easier to read/detect by consumers.
            raise InactiveRpcError(
                "EQC fetch_results failed due to grpc._channel._InactiveRpcError."
            ) from exc
        result = message_to_dict(eqc_results)
        # need to interpret result with correct precision
        # grpc serialization deserialization causes a change in the values
        # in order to ensure that results can be dumped to json
        # must use native python types if use float(np.float32(num))
        # then will get corrupted bits so must cast to str first
        if len(result["solution"]) > 0:
            result["solution"] = np.array(result["solution"])
            result["solution"] = [
                [float(f"{np.float32(val):.7f}") for val in soln]
                for soln in result["solution"]
            ]
        if len(result["distilled_solution"]) > 0:
            result["distilled_solution"] = np.array(result["distilled_solution"])

            result["distilled_solution"] = [
                [float(f"{np.float32(val):.7f}") for val in soln]
                for soln in result["distilled_solution"]
            ]
        result["energy"] = [
            float(f"{np.float32(energy):.7f}") for energy in result["energy"]
        ]
        result["distilled_energy"] = [
            float(f"{np.float32(energy):.7f}") for energy in result["distilled_energy"]
        ]
        return result

    def solve_sum_constrained(  # pylint: disable=R0913
        self,
        poly_coefficients: Union[List[float], np_typing.NDArray[np.float32]],
        poly_indices: Union[List[List[int]], np_typing.NDArray[np.uint32]],
        num_variables: Optional[int] = None,
        num_samples: int = 1,
        lock_id: str = "",
        relaxation_schedule: int = 1,
        sum_constraint: Union[int, float] = 10000,
        solution_precision: Optional[float] = None,
        mean_photon_number: Optional[float] = None,
        quantum_fluctuation_coefficient: Optional[int] = None,
    ) -> dict:
        """
        Utilizes Dirac to optimize a polynomial under the constraint that the
        sum of the solution values must equal the input :code:`sum_constraint`.

        :param poly_coefficients:
            the coefficient values for polynomial to be minimized. Numbers, including
            integers, should be floats with 32-bit (or less) precision, otherwise
            precision is lost during conversion to 32-bit.
        :param poly_indices:
            the indices for coefficient values for polynomial to be minimized.
        :param num_variables: optional input to specify number of variables for
            polynomial. Must be greater than or equal to maximum index value in
            :code:`poly_indices`.
        :param num_samples: the number of times to solve the problem on the device.
        :param lock_id: a UUID to allow for multi-user processing
        :param relaxation_schedule: four different schedules represented
            in integer parameter. Higher values reduce the variation in
            the analog spin values and therefore, are more probable to lead to
            improved objective function energy for input problem.
            Accepts range of values in set {1, 2, 3, 4}.
        :param sum_constraint: a normalization constraint that is applied to the
            problem space that is used to calculate :code:`energy`. This
            parameter will be rounded if exceeds float32 precision
            (e.g. 7-decimal places). Value must be between 1 and 10000.
        :param solution_precision: the level of precision to apply to the solutions.
            This parameter will be rounded if exceeds float32 precision
            (e.g. 7-decimal places). If specified a distillation method is
            applied to the continuous solutions to map them to the submitted
            :code:`solution_precision`. Input :code:`solution_precision` must
            satisfy :code:`solution_precision` greater than or equal to
            :code:`sum_constraint`/10000 in order to be valid.
            Also :code:`sum_constraint` must be divisible by :code:`solution_precision`.
            If :code:`solution_precision` is not specified no distillation will be
            applied to the solution derived by the device.
        :param mean_photon_number: optional parameter that modfies device
            configuration from the defaults for :code:`relaxation_schedule`.
            Sets the average number of photons that are present in a given
            quantum state.
        :param quantum_fluctuation_coefficient:
            Optional, overrides the default value from :code:`relaxation_schedule`. Specifies the
            number of photons (:math:`N`) in each feedback loop, which determines
            the shot noise. Accepts integer values in the range [0, 100].
            Shot noise scales inversely with the square root of the photon number:

        .. math::
            \\delta x \\propto \\frac{1}{\\sqrt{N}}

        :return: dict of results and timings with the following keys:

           - results: :class:`.SumConstrainedResult` dict
           - start_job_ts: time in ns marking start of job_submission
           - end_job_ts: time in ns marking end of job submission complete

        :note: The difference between `end_job_ts` and `start_job_ts` includes a 1
           second polling time which is not part of device solving time. To calculate
           the execution time on the device add together `preprocessing_time`, `runtime`
           and `postprocessing_time` from results object samplewise.
        """
        start_job = time.time_ns()
        submit_sum_constrained_job_resp = self.submit_sum_constrained_job(
            poly_coefficients=poly_coefficients,
            poly_indices=poly_indices,
            num_variables=num_variables,
            num_samples=num_samples,
            lock_id=lock_id,
            sum_constraint=sum_constraint,
            relaxation_schedule=relaxation_schedule,
            solution_precision=solution_precision,
            mean_photon_number=mean_photon_number,
            quantum_fluctuation_coefficient=quantum_fluctuation_coefficient,
        )
        if submit_sum_constrained_job_resp["err_code"] != 0:
            err_msg = (
                f"Submission failed with response: {submit_sum_constrained_job_resp}"
            )
            logging.error(err_msg, exc_info=True)
            raise RuntimeError(err_msg)
        sys_code = self.system_status()["status_code"]
        while sys_code != SysStatus.IDLE["status_code"]:
            sys_code = self.system_status()["status_code"]
            # this is based on the error statuses are 3 and above
            if sys_code > 3:
                err_msg = f"System unavailable status_code: {sys_code}"
                logging.error(err_msg, exc_info=True)
                raise RuntimeError(err_msg)
            # only sleep if not idle
            if sys_code != SysStatus.IDLE["status_code"]:
                time.sleep(1)
        # pull in results after is idle
        logging.info("Fetching sum-constrained results")
        job_result = self.fetch_sum_constrained_result(lock_id=lock_id)
        end_job = time.time_ns()
        if job_result["err_code"] != 0:
            raise RuntimeError(
                f"Job execution error\n"
                f"err_code: {job_result['err_code']}\n"
                f"err_desc: {job_result['err_desc']}"
            )
        job_result["start_job_ts"] = start_job
        job_result["end_job_ts"] = end_job
        return job_result

    def system_status(self) -> dict:
        """
        Client call to obtain EQC system status

        :returns: a member of :class:`eqc_direct.utils.SysStatus` as a dict:

            - **status_code**: `int`- current system status code
            - **status_desc**: `str`- description of current system status
        """
        try:
            sys_resp = self.eqc_stub.SystemStatus(eqc_pb2.Empty())
        except _InactiveRpcError as exc:
            raise InactiveRpcError(
                "EQC system_status failed due to grpc._channel._InactiveRpcError."
            ) from exc
        return message_to_dict(sys_resp)

    def acquire_lock(self) -> dict:
        """
        Makes a single attempt to acquire exclusive lock on hardware execution.
        Locking can be used to ensure orderly processing in multi-user environments.
        Lock can only be acquired when no other user has acquired the lock or when
        the system has been idle for 60 seconds while another user has the lock.
        This idle timeout prevents one user from blocking other users from using
        the machine even if they are not active.

        :return:
           a member of :class:`eqc_direct.utils.LockManageStatus` as a dict along
           with an additional key :code:`lock_id`:

           - **lock_id**: `str`- if acquired the current device `lock_id`
             else empty string
           - **status_code**: `int`- status code for lock id acquisition
           - **status_desc**: `str`- a description for the associated status code
        """
        try:
            acquire_lock_resp = self.eqc_stub.AcquireLock(eqc_pb2.Empty())
        except _InactiveRpcError as exc:
            raise InactiveRpcError(
                "EQC acquire_lock failed due to grpc._channel._InactiveRpcError."
            ) from exc

        return {
            "lock_id": acquire_lock_resp.lock_id,
            "status_code": acquire_lock_resp.lock_status.status_code,
            "status_desc": acquire_lock_resp.lock_status.status_desc,
        }

    def release_lock(self, lock_id: str = "") -> dict:
        """
        Releases exclusive lock for running health check or submitting job

        :param lock_id: a UUID with currently acquired exclusive device lock
        :return: a member of :class:`eqc_direct.utils.LockManageStatus` as a dict:

           - **status_code**: `int`- status code for lock id acquisition
           - **status_desc**: `str`- a description for the associated status code
        """
        release_input = eqc_pb2.LockMessage(lock_id=lock_id)
        try:
            release_lock_resp = self.eqc_stub.ReleaseLock(release_input)
        except _InactiveRpcError as exc:
            raise InactiveRpcError(
                "EQC release_lock failed due to grpc._channel._InactiveRpcError."
            ) from exc
        return message_to_dict(release_lock_resp)

    def check_lock(self, lock_id: str = "") -> dict:
        """
        Checks if submitted :code:`lock_id` has execution lock on the device

        :param lock_id: a UUID which will be checked to determine if has exclusive
            device execution lock
        :return: a member of :class:`eqc_direct.utils.LockCheckStatus` as a dict:

           - **status_code**: `int`- status code for lock check
           - **status_desc**: `str`- a description for the associated status code
        """
        check_input = eqc_pb2.LockMessage(lock_id=lock_id)
        check_output = self.eqc_stub.CheckLock(check_input)
        return message_to_dict(check_output)

    def stop_running_process(self, lock_id: str = "") -> dict:
        """
        Stops a running process either a health check or a Eqc job.
        Process locks will release automatically based on a timeout
        which is maintained in the server code if they are
        not released using this.

        :param lock_id: requires a lock_id that was acquired by
        :return:
           a member of :class:`eqc_direct.utils.SysStatus`
           as dict with following keys:

           - **status_code**: `int`- the system code after stopping
           - **status_desc**: `str`- the associated system status description
        """
        stop_input = eqc_pb2.LockMessage(lock_id=lock_id)
        try:
            stop_resp = self.eqc_stub.StopRunning(stop_input)
        except _InactiveRpcError as exc:
            raise InactiveRpcError(
                "EQC stop_running_process failed due to "
                "grpc._channel._InactiveRpcError."
            ) from exc
        return message_to_dict(stop_resp)

    def wait_for_lock(self) -> tuple:
        """
        Waits for lock indefinitely calling :func:`acquire_lock`

        :return: a tuple of the following items:

           - **lock_id**: `str`- exclusive lock for device execution with a timeout
           - **start_queue_ts**: `int`- time in ns on which lock was acquired is an int
           - **end_queue_ts**: `int`- time in ns on which queue for
             lock ended is an int.
        """
        lock_id = ""
        start_queue_ts = time.time_ns()
        while lock_id == "":
            sys_code = self.system_status()["status_code"]
            # this is based on the error statuses are 3 and above
            if sys_code >= 3:
                raise RuntimeError(f"System unavailable status_code: {sys_code}")
            lock_id = self.acquire_lock()["lock_id"]
            # only sleep if didn't get lock on device
            if lock_id == "":
                time.sleep(1)
        end_queue_ts = time.time_ns()
        return lock_id, start_queue_ts, end_queue_ts

    def system_info(self) -> SystemInfo:
        """
        Provides information regarding Dirac system

        :return: a :class:`eqc_direct.utils.SystemInfo` dict with a these items:

            - **server_version**: `str` - the gRPC server version
            - **device_type**: `str` - the device type (e.g., Dirac-3)
            - **fpga_version**: `str` - version of FPGA in device (None if using the simulator)
        """
        try:
            return SystemInfo(
                **message_to_dict(self.eqc_stub.SystemInfo(eqc_pb2.Empty()))
            )
        except _InactiveRpcError as exc:
            raise InactiveRpcError(
                "EQC server_info call failed due to inactive grpc channel"
            ) from exc



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from eqc_direct import eqc_pb2 as eqc__direct_dot_eqc__pb2


class EqcServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.SolveSumConst = channel.unary_unary(
            "/eqc.EqcService/SolveSumConst",
            request_serializer=eqc__direct_dot_eqc__pb2.SumConstInput.SerializeToString,
            response_deserializer=eqc__direct_dot_eqc__pb2.SubmitOutput.FromString,
        )
        self.SolveInt = channel.unary_unary(
            "/eqc.EqcService/SolveInt",
            request_serializer=eqc__direct_dot_eqc__pb2.IntInput.SerializeToString,
            response_deserializer=eqc__direct_dot_eqc__pb2.SubmitOutput.FromString,
        )
        self.FetchIntResults = channel.unary_unary(
            "/eqc.EqcService/FetchIntResults",
            request_serializer=eqc__direct_dot_eqc__pb2.LockMessage.SerializeToString,
            response_deserializer=eqc__direct_dot_eqc__pb2.IntResult.FromString,
        )
        self.FetchSumConstResults = channel.unary_unary(
            "/eqc.EqcService/FetchSumConstResults",
            request_serializer=eqc__direct_dot_eqc__pb2.LockMessage.SerializeToString,
            response_deserializer=eqc__direct_dot_eqc__pb2.SumConstResult.FromString,
        )
        self.SystemStatus = channel.unary_unary(
            "/eqc.EqcService/SystemStatus",
            request_serializer=eqc__direct_dot_eqc__pb2.Empty.SerializeToString,
            response_deserializer=eqc__direct_dot_eqc__pb2.StatusOutput.FromString,
        )
        self.StopRunning = channel.unary_unary(
            "/eqc.EqcService/StopRunning",
            request_serializer=eqc__direct_dot_eqc__pb2.LockMessage.SerializeToString,
            response_deserializer=eqc__direct_dot_eqc__pb2.StatusOutput.FromString,
        )
        self.AcquireLock = channel.unary_unary(
            "/eqc.EqcService/AcquireLock",
            request_serializer=eqc__direct_dot_eqc__pb2.Empty.SerializeToString,
            response_deserializer=eqc__direct_dot_eqc__pb2.LockOutput.FromString,
        )
        self.ReleaseLock = channel.unary_unary(
            "/eqc.EqcService/ReleaseLock",
            request_serializer=eqc__direct_dot_eqc__pb2.LockMessage.SerializeToString,
            response_deserializer=eqc__direct_dot_eqc__pb2.StatusOutput.FromString,
        )
        self.CheckLock = channel.unary_unary(
            "/eqc.EqcService/CheckLock",
            request_serializer=eqc__direct_dot_eqc__pb2.LockMessage.SerializeToString,
            response_deserializer=eqc__direct_dot_eqc__pb2.StatusOutput.FromString,
        )
        self.SystemInfo = channel.unary_unary(
            "/eqc.EqcService/SystemInfo",
            request_serializer=eqc__direct_dot_eqc__pb2.Empty.SerializeToString,
            response_deserializer=eqc__direct_dot_eqc__pb2.VersionOutput.FromString,
        )
        self.MixedIntExp = channel.unary_unary(
            "/eqc.EqcService/MixedIntExp",
            request_serializer=eqc__direct_dot_eqc__pb2.MixedIntExpInput.SerializeToString,
            response_deserializer=eqc__direct_dot_eqc__pb2.SubmitOutput.FromString,
        )
        self.FetchMixedIntExpResults = channel.unary_unary(
            "/eqc.EqcService/FetchMixedIntExpResults",
            request_serializer=eqc__direct_dot_eqc__pb2.LockMessage.SerializeToString,
            response_deserializer=eqc__direct_dot_eqc__pb2.MixedIntExpResult.FromString,
        )
        self.SolveSumConstExp = channel.unary_unary(
            "/eqc.EqcService/SolveSumConstExp",
            request_serializer=eqc__direct_dot_eqc__pb2.SumConstExpInput.SerializeToString,
            response_deserializer=eqc__direct_dot_eqc__pb2.SubmitOutput.FromString,
        )
        self.FetchSumConstExpResults = channel.unary_unary(
            "/eqc.EqcService/FetchSumConstExpResults",
            request_serializer=eqc__direct_dot_eqc__pb2.LockMessage.SerializeToString,
            response_deserializer=eqc__direct_dot_eqc__pb2.SumConstExpResult.FromString,
        )
        self.GetPhotonRate = channel.unary_unary(
            "/eqc.EqcService/GetPhotonRate",
            request_serializer=eqc__direct_dot_eqc__pb2.RateInput.SerializeToString,
            response_deserializer=eqc__direct_dot_eqc__pb2.RateOutput.FromString,
        )


class EqcServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def SolveSumConst(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def SolveInt(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def FetchIntResults(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def FetchSumConstResults(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def SystemStatus(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def StopRunning(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def AcquireLock(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ReleaseLock(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def CheckLock(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def SystemInfo(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def MixedIntExp(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def FetchMixedIntExpResults(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def SolveSumConstExp(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def FetchSumConstExpResults(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def GetPhotonRate(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")


def add_EqcServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
        "SolveSumConst": grpc.unary_unary_rpc_method_handler(
            servicer.SolveSumConst,
            request_deserializer=eqc__direct_dot_eqc__pb2.SumConstInput.FromString,
            response_serializer=eqc__direct_dot_eqc__pb2.SubmitOutput.SerializeToString,
        ),
        "SolveInt": grpc.unary_unary_rpc_method_handler(
            servicer.SolveInt,
            request_deserializer=eqc__direct_dot_eqc__pb2.IntInput.FromString,
            response_serializer=eqc__direct_dot_eqc__pb2.SubmitOutput.SerializeToString,
        ),
        "FetchIntResults": grpc.unary_unary_rpc_method_handler(
            servicer.FetchIntResults,
            request_deserializer=eqc__direct_dot_eqc__pb2.LockMessage.FromString,
            response_serializer=eqc__direct_dot_eqc__pb2.IntResult.SerializeToString,
        ),
        "FetchSumConstResults": grpc.unary_unary_rpc_method_handler(
            servicer.FetchSumConstResults,
            request_deserializer=eqc__direct_dot_eqc__pb2.LockMessage.FromString,
            response_serializer=eqc__direct_dot_eqc__pb2.SumConstResult.SerializeToString,
        ),
        "SystemStatus": grpc.unary_unary_rpc_method_handler(
            servicer.SystemStatus,
            request_deserializer=eqc__direct_dot_eqc__pb2.Empty.FromString,
            response_serializer=eqc__direct_dot_eqc__pb2.StatusOutput.SerializeToString,
        ),
        "StopRunning": grpc.unary_unary_rpc_method_handler(
            servicer.StopRunning,
            request_deserializer=eqc__direct_dot_eqc__pb2.LockMessage.FromString,
            response_serializer=eqc__direct_dot_eqc__pb2.StatusOutput.SerializeToString,
        ),
        "AcquireLock": grpc.unary_unary_rpc_method_handler(
            servicer.AcquireLock,
            request_deserializer=eqc__direct_dot_eqc__pb2.Empty.FromString,
            response_serializer=eqc__direct_dot_eqc__pb2.LockOutput.SerializeToString,
        ),
        "ReleaseLock": grpc.unary_unary_rpc_method_handler(
            servicer.ReleaseLock,
            request_deserializer=eqc__direct_dot_eqc__pb2.LockMessage.FromString,
            response_serializer=eqc__direct_dot_eqc__pb2.StatusOutput.SerializeToString,
        ),
        "CheckLock": grpc.unary_unary_rpc_method_handler(
            servicer.CheckLock,
            request_deserializer=eqc__direct_dot_eqc__pb2.LockMessage.FromString,
            response_serializer=eqc__direct_dot_eqc__pb2.StatusOutput.SerializeToString,
        ),
        "SystemInfo": grpc.unary_unary_rpc_method_handler(
            servicer.SystemInfo,
            request_deserializer=eqc__direct_dot_eqc__pb2.Empty.FromString,
            response_serializer=eqc__direct_dot_eqc__pb2.VersionOutput.SerializeToString,
        ),
        "MixedIntExp": grpc.unary_unary_rpc_method_handler(
            servicer.MixedIntExp,
            request_deserializer=eqc__direct_dot_eqc__pb2.MixedIntExpInput.FromString,
            response_serializer=eqc__direct_dot_eqc__pb2.SubmitOutput.SerializeToString,
        ),
        "FetchMixedIntExpResults": grpc.unary_unary_rpc_method_handler(
            servicer.FetchMixedIntExpResults,
            request_deserializer=eqc__direct_dot_eqc__pb2.LockMessage.FromString,
            response_serializer=eqc__direct_dot_eqc__pb2.MixedIntExpResult.SerializeToString,
        ),
        "SolveSumConstExp": grpc.unary_unary_rpc_method_handler(
            servicer.SolveSumConstExp,
            request_deserializer=eqc__direct_dot_eqc__pb2.SumConstExpInput.FromString,
            response_serializer=eqc__direct_dot_eqc__pb2.SubmitOutput.SerializeToString,
        ),
        "FetchSumConstExpResults": grpc.unary_unary_rpc_method_handler(
            servicer.FetchSumConstExpResults,
            request_deserializer=eqc__direct_dot_eqc__pb2.LockMessage.FromString,
            response_serializer=eqc__direct_dot_eqc__pb2.SumConstExpResult.SerializeToString,
        ),
        "GetPhotonRate": grpc.unary_unary_rpc_method_handler(
            servicer.GetPhotonRate,
            request_deserializer=eqc__direct_dot_eqc__pb2.RateInput.FromString,
            response_serializer=eqc__direct_dot_eqc__pb2.RateOutput.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
        "eqc.EqcService", rpc_method_handlers
    )
    server.add_generic_rpc_handlers((generic_handler,))


# This class is part of an EXPERIMENTAL API.
class EqcService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def SolveSumConst(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/eqc.EqcService/SolveSumConst",
            eqc__direct_dot_eqc__pb2.SumConstInput.SerializeToString,
            eqc__direct_dot_eqc__pb2.SubmitOutput.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def SolveInt(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/eqc.EqcService/SolveInt",
            eqc__direct_dot_eqc__pb2.IntInput.SerializeToString,
            eqc__direct_dot_eqc__pb2.SubmitOutput.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def FetchIntResults(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/eqc.EqcService/FetchIntResults",
            eqc__direct_dot_eqc__pb2.LockMessage.SerializeToString,
            eqc__direct_dot_eqc__pb2.IntResult.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def FetchSumConstResults(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/eqc.EqcService/FetchSumConstResults",
            eqc__direct_dot_eqc__pb2.LockMessage.SerializeToString,
            eqc__direct_dot_eqc__pb2.SumConstResult.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def SystemStatus(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/eqc.EqcService/SystemStatus",
            eqc__direct_dot_eqc__pb2.Empty.SerializeToString,
            eqc__direct_dot_eqc__pb2.StatusOutput.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def StopRunning(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/eqc.EqcService/StopRunning",
            eqc__direct_dot_eqc__pb2.LockMessage.SerializeToString,
            eqc__direct_dot_eqc__pb2.StatusOutput.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def AcquireLock(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/eqc.EqcService/AcquireLock",
            eqc__direct_dot_eqc__pb2.Empty.SerializeToString,
            eqc__direct_dot_eqc__pb2.LockOutput.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def ReleaseLock(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/eqc.EqcService/ReleaseLock",
            eqc__direct_dot_eqc__pb2.LockMessage.SerializeToString,
            eqc__direct_dot_eqc__pb2.StatusOutput.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def CheckLock(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/eqc.EqcService/CheckLock",
            eqc__direct_dot_eqc__pb2.LockMessage.SerializeToString,
            eqc__direct_dot_eqc__pb2.StatusOutput.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def SystemInfo(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/eqc.EqcService/SystemInfo",
            eqc__direct_dot_eqc__pb2.Empty.SerializeToString,
            eqc__direct_dot_eqc__pb2.VersionOutput.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def MixedIntExp(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/eqc.EqcService/MixedIntExp",
            eqc__direct_dot_eqc__pb2.MixedIntExpInput.SerializeToString,
            eqc__direct_dot_eqc__pb2.SubmitOutput.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def FetchMixedIntExpResults(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/eqc.EqcService/FetchMixedIntExpResults",
            eqc__direct_dot_eqc__pb2.LockMessage.SerializeToString,
            eqc__direct_dot_eqc__pb2.MixedIntExpResult.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def SolveSumConstExp(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/eqc.EqcService/SolveSumConstExp",
            eqc__direct_dot_eqc__pb2.SumConstExpInput.SerializeToString,
            eqc__direct_dot_eqc__pb2.SubmitOutput.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def FetchSumConstExpResults(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/eqc.EqcService/FetchSumConstExpResults",
            eqc__direct_dot_eqc__pb2.LockMessage.SerializeToString,
            eqc__direct_dot_eqc__pb2.SumConstExpResult.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def GetPhotonRate(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/eqc.EqcService/GetPhotonRate",
            eqc__direct_dot_eqc__pb2.RateInput.SerializeToString,
            eqc__direct_dot_eqc__pb2.RateOutput.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )



"""
Package initialization.

Copyright 2024, Quantum Computing Incorporated
"""

import importlib.metadata
import os

PACKAGE_NAME = __name__
PACKAGE_VERSION = importlib.metadata.version(PACKAGE_NAME)

DEVICE_IP_ADDRESS_DEFAULT = os.getenv("DEVICE_IP_ADDRESS", "localhost")
DEVICE_PORT_DEFAULT = os.getenv("DEVICE_PORT", "50051")



import time
import os
import sys
import logging

# sys.argv[0] is the python script
ip_address = sys.argv[1]
port = sys.argv[2]
log_path = sys.argv[3]

if not os.path.exists(log_path):
    os.makedirs(log_path)
log_file = log_path + "/server_" + time.strftime("%Y%m%d-%H%M%S") + ".log"
logging.basicConfig(
    filename=log_file,
    force=True,
    format=(
        "%(levelname) -10s %(asctime)s %(name) -30s %(funcName) "
        "-35s %(lineno) -5d: %(message)s"
    ),
    filemode="a",
    datefmt="%H:%M:%S",
    level=logging.INFO,
)

rootLogger = logging.getLogger()
consoleHandler = logging.StreamHandler()
rootLogger.addHandler(consoleHandler)

rootLogger.info("Logging to file: %s", log_file)

# https://betterstack.com/community/questions/how-to-log-uncaught-exceptions-in-python/
# Creating a handler
def handle_unhandled_exception(exc_type, exc_value, exc_traceback):
    if issubclass(exc_type, KeyboardInterrupt):
        # Will call default excepthook
        sys.__excepthook__(exc_type, exc_value, exc_traceback)
        return
        # Create a critical level log message with info from the except hook.
    logging.critical(
        "Unhandled exception", exc_info=(exc_type, exc_value, exc_traceback)
    )


# Assign the excepthook to the handler
sys.excepthook = handle_unhandled_exception
if __name__ == "__main__":
    from eqc_direct.server_sim import GrpcServer

    # runs on ThreadPoolExecutor
    server_cls = GrpcServer(
        ip_address=ip_address,
        port=port,
        #       private_key="private.pem",
        #       cert_file="certificate.pem",
    )
    server_cls.serve()
    server_cls.join()



"""
Relies on environment var1 DEVICE_IP_ADDRESS to identify IP_ADDRESS FOR tests
"""
import os
import time
import unittest
import warnings
import sys

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

import numpy as np
from grpc_server_cls import ServerForTest
from eqc_direct.client import EqcClient
from eqc_direct.utils import JobCodes, LockCheckStatus, LockManageStatus


class TestEqcClient(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.test_ip_addr = os.getenv("DEVICE_IP_ADDRESS", "localhost")
        # only instantiate server if running on localhost
        if cls.test_ip_addr == "localhost":
            cls.server = ServerForTest(ip_address=cls.test_ip_addr, port="50051")
            cls.server.start()
        print("Running test using ip_address", cls.test_ip_addr)
        cls.eqc_client = EqcClient(ip_address=cls.test_ip_addr)
        cls.poly_indices = np.array(
            [
                [0, 0, 0, 3],
                [0, 0, 2, 2],
                [0, 0, 2, 3],
                [0, 1, 2, 3],
                [0, 1, 1, 1],
                [0, 2, 2, 3],
                [3, 3, 3, 3],
                [1, 1, 2, 3],
            ],
        )
        cls.poly_coefficients = np.array([1, 2, 3, 4, 5, 6, 7, 8], dtype=np.float32)
        cls.num_variables = 3
        cls.num_samples = 7

    @classmethod
    def tearDownClass(cls):
        if cls.test_ip_addr == "localhost":
            cls.server.stop()

    def test_ip_address(self):
        self.assertEqual(
            self.eqc_client.ip_address,
            os.getenv("DEVICE_IP_ADDRESS", "localhost"),
        )

    def test_port(self):
        self.assertEqual(self.eqc_client.port, "50051")

    def test_ip_add_port(self):
        self.assertEqual(
            self.eqc_client.ip_add_port,
            f"{self.eqc_client.ip_address}:{self.eqc_client.port}",
        )

    def test_numpy_dtype_preserved_in_broadcast_to(self):
        self.assertEqual(
            np.broadcast_to(np.array(1, dtype=np.float32), (self.num_variables,)).dtype,
            np.float32,
        )

    def test_numpy_dtype_preserved_in_array_cast(self):
        self.assertEqual(np.array(np.array([1, 2], dtype=np.float32)).dtype, np.float32)

    def test_check_lock(self):
        try:
            lock_av = self.eqc_client.check_lock()
            print("lock_av", lock_av)
            self.assertDictEqual(lock_av, LockCheckStatus.AVAILABLE)
            lock_id = self.eqc_client.acquire_lock()["lock_id"]
            lock_unv = self.eqc_client.check_lock()
            print("lock_unv", lock_unv)
            self.assertDictEqual(lock_unv, LockCheckStatus.UNAVAILABLE)
            lock_user = self.eqc_client.check_lock(lock_id=lock_id)
            self.assertDictEqual(lock_user, LockCheckStatus.USER_LOCKED)
        finally:
            self.eqc_client.release_lock(lock_id=lock_id)

    def test_basic_locking(self):
        try:
            print("Test basic locking")
            succ_lock_resp = self.eqc_client.acquire_lock()
            lock_id = succ_lock_resp["lock_id"]
            self.assertIsInstance(lock_id, str)
            self.assertEqual(36, len(lock_id))
            lock_fail = self.eqc_client.acquire_lock()
            self.assertEqual(lock_fail["lock_id"], "")

            self.assertEqual(lock_fail["status_code"], 2)
            self.assertEqual(
                lock_fail["status_desc"],
                "Lock currently in use unable to perform operation",
            )
        finally:
            release_out = self.eqc_client.release_lock(lock_id=lock_id)
            self.assertEqual(release_out["status_code"], 0)
            self.assertEqual(release_out["status_desc"], "Success")

    def test_wait_for_lock(self):
        print("Test wait for lock")
        try:
            lock_id, start_ts, end_ts = self.eqc_client.wait_for_lock()
            self.assertIsInstance(lock_id, str)
            self.assertEqual(36, len(lock_id))
            self.assertTrue(start_ts < end_ts)
            self.assertIsInstance(start_ts, int)
            self.assertIsInstance(end_ts, int)
        finally:
            self.eqc_client.release_lock(lock_id=lock_id)

    def test_stop_running_process(self):
        try:
            print("Stop test")
            lock_id, _, _ = self.eqc_client.wait_for_lock()
            # this isn't a real problem since isn't symmetric
            # just chose one that is large enough to take a significant
            # time to do data transfer so can cancel.
            print("stop lock", lock_id)
            submit_output = self.eqc_client.submit_sum_constrained_job(
                poly_indices=self.poly_indices,
                poly_coefficients=self.poly_coefficients,
                num_variables=self.num_variables,
                relaxation_schedule=1,
                sum_constraint=40,
                lock_id=lock_id,
            )
            print("SUBMTI OUTPUT", submit_output)
            stop_out = self.eqc_client.stop_running_process(lock_id=lock_id)
            self.assertDictEqual(stop_out, LockManageStatus.SUCCESS)
            stop_res = self.eqc_client.fetch_sum_constrained_result(lock_id=lock_id)
            self.assertEqual(stop_res["err_code"], JobCodes.USER_INTERRUPT["err_code"])
            self.assertEqual(stop_res["err_desc"], JobCodes.USER_INTERRUPT["err_desc"])

        finally:
            self.eqc_client.release_lock(lock_id=lock_id)

    def test_stop_lock_mismatch(self):
        stop_output = self.eqc_client.stop_running_process(lock_id="bad_lock")
        self.assertDictEqual(stop_output, LockManageStatus.MISMATCH)

    def test_system_info(self):
        system_info_resp = self.eqc_client.system_info()
        self.assertIsInstance(system_info_resp["server_version"], str)
        self.assertIsInstance(system_info_resp["device_type"], str)
        self.assertIsInstance(system_info_resp["fpga_version"], str)

    def test_system_status(self):
        sys_resp = self.eqc_client.system_status()
        self.assertListEqual(list(sys_resp.keys()), ["status_code", "status_desc"])
        self.assertIsInstance(sys_resp["status_code"], int)
        self.assertIsInstance(sys_resp["status_desc"], str)

    def test_submit_exceeds_size(self):
        print("test submit job sum constraint")
        lock_id, _, _ = self.eqc_client.wait_for_lock()
        # degree 2
        poly_indices_2 = np.array([[0, 950]])
        poly_coefficients = np.array([10], dtype=np.float32)
        try:
            print("DEGREE 2")
            result = self.eqc_client.submit_sum_constrained_job(
                poly_indices=poly_indices_2,
                poly_coefficients=poly_coefficients,
                relaxation_schedule=1,
                sum_constraint=100,
                lock_id=lock_id,
            )
            print("Sum constraint:", result)
            self.assertDictEqual(
                result,
                JobCodes.EXCEEDS_MAX_SIZE,
            )
            print("DEGREE 3")
            poly_indices_3 = [[0, 0, 136]]
            result = self.eqc_client.submit_sum_constrained_job(
                poly_indices=np.array(poly_indices_3),
                poly_coefficients=poly_coefficients,
                relaxation_schedule=1,
                sum_constraint=100,
                lock_id=lock_id,
            )
            print("Sum constraint:", result)
            self.assertDictEqual(
                result,
                JobCodes.EXCEEDS_MAX_SIZE,
            )
            print("DEGREE 4")
            poly_indices_4 = np.array([[0, 0, 0, 40]])
            result = self.eqc_client.submit_sum_constrained_job(
                poly_indices=poly_indices_4,
                poly_coefficients=poly_coefficients,
                relaxation_schedule=1,
                sum_constraint=100,
                lock_id=lock_id,
            )
            print("Sum constraint:", result)
            self.assertDictEqual(
                result,
                JobCodes.EXCEEDS_MAX_SIZE,
            )
            print("DEGREE 5")
            poly_indices_5 = np.array([[0, 0, 0, 0, 20]])
            result = self.eqc_client.submit_sum_constrained_job(
                poly_indices=poly_indices_5,
                poly_coefficients=poly_coefficients,
                relaxation_schedule=1,
                sum_constraint=100,
                lock_id=lock_id,
            )
            print("Sum constraint:", result)
            self.assertDictEqual(
                result,
                JobCodes.EXCEEDS_MAX_SIZE,
            )
        finally:
            self.eqc_client.release_lock(lock_id=lock_id)

    def test_solve_sum_constrained_and_fetch_results_batched_samples_non_distilled(
        self,
    ):
        try:
            print("Process job tests with batched samples for non-distilled solution")
            lock_id, _, _ = self.eqc_client.wait_for_lock()
            solve_sum_constrained_output = self.eqc_client.solve_sum_constrained(
                poly_indices=self.poly_indices,
                poly_coefficients=self.poly_coefficients,
                relaxation_schedule=1,
                sum_constraint=40,
                lock_id=lock_id,
                num_samples=self.num_samples,
            )
            self.assertEqual(solve_sum_constrained_output["err_code"], 0)
            job_result = self.eqc_client.fetch_sum_constrained_result(lock_id=lock_id)
            self.assertEqual(job_result["err_code"], 0)
            self.assertEqual(job_result["err_desc"], "Success")
            self.assertEqual(job_result["num_samples"], self.num_samples)
            self.assertEqual(job_result["num_variables"], self.num_variables)
            self.assertTrue(job_result["preprocessing_time"] > 0)
            self.assertIsInstance(job_result["calibration_time"], float)
            self.assertTrue(len(job_result["postprocessing_time"]) == self.num_samples)
            self.assertTrue(len(job_result["runtime"]) == self.num_samples)
            self.assertTrue(len(job_result["energy"]) == self.num_samples)
            self.assertTrue(len(job_result["solution"]) == self.num_samples)
            self.assertTrue(len(job_result["distilled_energy"]) == 0)
            self.assertTrue(len(job_result["distilled_solution"]) == 0)
            for k in range(self.num_samples):
                self.assertTrue(job_result["runtime"][k] > 0)
                self.assertIsInstance(job_result["energy"][k], float)
                self.assertIsInstance(job_result["postprocessing_time"][k], float)
                self.assertTrue(len(job_result["solution"][k]), self.num_variables)
        finally:
            self.eqc_client.release_lock(lock_id=lock_id)

    def test_solve_sum_constrained_and_fetch_results_batched_samples_distilled(self):
        try:
            print("Process job tests with batched samples for distilled solution")
            lock_id, _, _ = self.eqc_client.wait_for_lock()
            solve_sum_constrained_output = self.eqc_client.solve_sum_constrained(
                poly_indices=self.poly_indices,
                poly_coefficients=self.poly_coefficients,
                relaxation_schedule=1,
                sum_constraint=40,
                solution_precision=0.1,
                lock_id=lock_id,
                num_samples=self.num_samples,
            )
            self.assertEqual(solve_sum_constrained_output["err_code"], 0)
            job_result = self.eqc_client.fetch_sum_constrained_result(lock_id=lock_id)
            print("BATCH RESULT:", job_result)
            self.assertEqual(job_result["err_code"], 0)
            self.assertEqual(job_result["err_desc"], "Success")
            self.assertEqual(job_result["num_samples"], self.num_samples)
            self.assertEqual(job_result["num_variables"], self.num_variables)
            self.assertTrue(job_result["preprocessing_time"] > 0)
            self.assertIsInstance(job_result["calibration_time"], float)
            self.assertTrue(len(job_result["postprocessing_time"]) == self.num_samples)
            self.assertTrue(len(job_result["runtime"]) == self.num_samples)
            self.assertTrue(len(job_result["energy"]) == self.num_samples)
            self.assertTrue(len(job_result["solution"]) == self.num_samples)
            self.assertTrue(len(job_result["distilled_energy"]) == self.num_samples)
            self.assertTrue(len(job_result["distilled_solution"]) == self.num_samples)
            for k in range(self.num_samples):
                self.assertTrue(job_result["runtime"][k] > 0)
                self.assertIsInstance(job_result["postprocessing_time"][k], float)
                self.assertIsInstance(job_result["energy"][k], float)
                self.assertTrue(len(job_result["solution"][k]), self.num_variables)
                self.assertIsInstance(job_result["distilled_energy"][k], float)
                self.assertTrue(
                    len(job_result["distilled_solution"][k]), self.num_variables
                )
        finally:
            self.eqc_client.release_lock(lock_id=lock_id)

    def test_submit_job_and_fetch_results(self):
        try:
            print("Submit job tests")
            lock_id, _, _ = self.eqc_client.wait_for_lock()
            submit_output = self.eqc_client.submit_sum_constrained_job(
                poly_indices=self.poly_indices,
                poly_coefficients=self.poly_coefficients,
                relaxation_schedule=1,
                sum_constraint=40,
                lock_id=lock_id,
            )
            self.assertEqual(submit_output["err_code"], 0)
            while self.eqc_client.system_status()["status_code"] != 0:
                time.sleep(1)
            job_result = self.eqc_client.fetch_sum_constrained_result(lock_id=lock_id)
            print(job_result)
            self.assertEqual(job_result["err_code"], 0)
            print("assert desc")
            print("JOB RESULT", job_result)
            self.assertEqual(job_result["err_desc"], "Success")
            print("assert run")
            self.assertTrue(job_result["runtime"][0] >= 0)
            self.assertIsInstance(job_result["energy"][0], float)
            print("solution length")
            self.assertTrue(len(job_result["solution"][0]), 3)
            bad_res = self.eqc_client.fetch_sum_constrained_result(lock_id="bad")
            self.assertDictEqual(
                {"err_code": bad_res["err_code"], "err_desc": bad_res["err_desc"]},
                JobCodes.LOCK_MISMATCH,
            )
        finally:
            self.eqc_client.release_lock(lock_id=lock_id)

    def test_solve_sum_constrained(self):
        print("Test solve_sum_constrained")
        try:
            lock_id, _, _ = self.eqc_client.wait_for_lock()
            result = self.eqc_client.solve_sum_constrained(
                poly_indices=self.poly_indices,
                poly_coefficients=self.poly_coefficients,
                relaxation_schedule=1,
                sum_constraint=40,
                lock_id=lock_id,
            )
            print(result)
            self.assertEqual(len(result["solution"][0]), np.max(self.poly_indices))
            self.assertEqual(len(result["distilled_solution"]), 0)
        finally:
            self.eqc_client.release_lock(lock_id=lock_id)

    def test_solve_sum_constrained_submit_fail(self):
        print("Test solve_sum_constrained with bad lock")
        with self.assertRaises(RuntimeError):
            self.eqc_client.solve_sum_constrained(
                poly_indices=self.poly_indices,
                poly_coefficients=self.poly_coefficients,
                relaxation_schedule=1,
                sum_constraint=40,
                lock_id="bad_lock",
            )

    def test_submit_job_sum_constraint(self):
        print("test submit job sum constraint")
        lock_id, _, _ = self.eqc_client.wait_for_lock()
        try:
            result = self.eqc_client.submit_sum_constrained_job(
                poly_indices=self.poly_indices,
                poly_coefficients=self.poly_coefficients,
                relaxation_schedule=1,
                sum_constraint=-40,
                lock_id=lock_id,
            )
            print("Sum constraint:", result)
            self.assertDictEqual(
                result,
                JobCodes.INVALID_SUM_CONSTRAINT,
            )
            result = self.eqc_client.submit_sum_constrained_job(
                poly_indices=self.poly_indices,
                poly_coefficients=self.poly_coefficients,
                relaxation_schedule=1,
                sum_constraint=10001,
                lock_id=lock_id,
            )
            print("Sum constraint:", result)
            self.assertDictEqual(
                result,
                JobCodes.INVALID_SUM_CONSTRAINT,
            )

        finally:
            self.eqc_client.release_lock(lock_id=lock_id)

    def test_submit_job_invalid_precision(self):
        print("INVALID PRECISION")
        lock_id, _, _ = self.eqc_client.wait_for_lock()
        try:
            result = self.eqc_client.submit_sum_constrained_job(
                poly_indices=self.poly_indices,
                poly_coefficients=self.poly_coefficients,
                relaxation_schedule=1,
                sum_constraint=9999,
                solution_precision=0.1,
                lock_id=lock_id,
            )
            print("INVALID_PREC_RESULT", result)
            self.assertDictEqual(
                result,
                JobCodes.INVALID_PRECISION,
            )
        finally:
            self.eqc_client.release_lock(lock_id=lock_id)

    def test_submit_job_precision(self):
        print("Test solve_sum_constrained precision")
        lock_id, _, _ = self.eqc_client.wait_for_lock()
        try:
            result = self.eqc_client.solve_sum_constrained(
                poly_indices=self.poly_indices,
                poly_coefficients=self.poly_coefficients,
                relaxation_schedule=1,
                sum_constraint=2.4,
                solution_precision=0.24,
                lock_id=lock_id,
            )
        finally:
            self.eqc_client.release_lock(lock_id=lock_id)

    def test_submit_job_precision_constraint_mismatch(self):
        print("Test constraint precision mismatch")
        lock_id, _, _ = self.eqc_client.wait_for_lock()
        try:
            result = self.eqc_client.submit_sum_constrained_job(
                poly_indices=self.poly_indices,
                poly_coefficients=self.poly_coefficients,
                relaxation_schedule=1,
                sum_constraint=1,
                solution_precision=0.11,
                lock_id=lock_id,
            )
            self.assertDictEqual(
                result,
                JobCodes.PRECISION_CONSTRAINT_MISMATCH,
            )
        finally:
            self.eqc_client.release_lock(lock_id=lock_id)

    def test_submit_job_decreasing_index(self):
        print("Test decreasing")
        lock_id, _, _ = self.eqc_client.wait_for_lock()
        try:
            result = self.eqc_client.submit_sum_constrained_job(
                poly_indices=np.array([[1, 0], [0, 1]]),
                poly_coefficients=np.array([1, 1], dtype=np.float32),
                relaxation_schedule=1,
                sum_constraint=1,
                solution_precision=0.1,
                lock_id=lock_id,
            )
            self.assertDictEqual(
                result,
                JobCodes.DECREASING_INDEX,
            )
        finally:
            self.eqc_client.release_lock(lock_id=lock_id)

    def test_warn_below_recommended_soln_precision(self):
        print("Precision warn below")
        high_prec_coefs = np.random.normal(size=8)
        lock_id, _, _ = self.eqc_client.wait_for_lock()
        with self.assertWarns(
            Warning,
            msg="Minimum recommended levels (sum_constraint/solution_precision) is 200",
        ):
            try:
                result = self.eqc_client.solve_sum_constrained(
                    poly_indices=self.poly_indices,
                    poly_coefficients=high_prec_coefs,
                    relaxation_schedule=1,
                    sum_constraint=100,
                    solution_precision=1,
                    lock_id=lock_id,
                )
            finally:
                self.eqc_client.release_lock(lock_id=lock_id)

    def test_precision_warning(self):
        print("Precision warn")
        high_prec_coefs = np.random.normal(size=8)
        lock_id, _, _ = self.eqc_client.wait_for_lock()
        with self.assertWarns(
            Warning,
            msg="Max precision for EQC device is float32 input type was float64. Input matrix will be rounded",
        ):
            try:
                result = self.eqc_client.solve_sum_constrained(
                    poly_indices=self.poly_indices,
                    poly_coefficients=high_prec_coefs,
                    relaxation_schedule=1,
                    sum_constraint=200,
                    solution_precision=1,
                    lock_id=lock_id,
                )
            finally:
                self.eqc_client.release_lock(lock_id=lock_id)

    def test_integer_conversion_soln(self):
        print("Test Integer")
        try:
            lock_id, _, _ = self.eqc_client.wait_for_lock()
            result = self.eqc_client.solve_sum_constrained(
                poly_indices=self.poly_indices,
                poly_coefficients=self.poly_coefficients,
                relaxation_schedule=1,
                sum_constraint=100,
                lock_id=lock_id,
                solution_precision=1,
            )
            print(result)
            all([val.is_integer() for val in result["solution"][0]])
        finally:
            self.eqc_client.release_lock(lock_id=lock_id)

    def test_precision_soln(self):
        print("Test Continuous Solution")
        try:
            lock_id, _, _ = self.eqc_client.wait_for_lock()
            prec_val = 0.01
            result = self.eqc_client.solve_sum_constrained(
                poly_indices=self.poly_indices,
                poly_coefficients=self.poly_coefficients,
                relaxation_schedule=1,
                sum_constraint=100,
                lock_id=lock_id,
                solution_precision=prec_val,
            )
            print(result)
            print(result["distilled_solution"])
            all([(val / prec_val).is_integer() for val in result["solution"][0]])
        finally:
            self.eqc_client.release_lock(lock_id=lock_id)

    def test_invalid_schedule(self):
        print("invalid schedule")
        try:
            lock_id, _, _ = self.eqc_client.wait_for_lock()
            result = self.eqc_client.submit_sum_constrained_job(
                poly_indices=self.poly_indices,
                poly_coefficients=self.poly_coefficients,
                relaxation_schedule=10,
                sum_constraint=100,
                lock_id=lock_id,
                solution_precision=1,
            )
            self.assertDictEqual(result, JobCodes.INVALID_RELAXATION_SCHEDULE)
        finally:
            self.eqc_client.release_lock(lock_id=lock_id)

    def test_lock_mismatch(self):
        print("lock_mismatch")
        try:
            lock_id, _, _ = self.eqc_client.wait_for_lock()
            result = self.eqc_client.submit_sum_constrained_job(
                poly_indices=self.poly_indices,
                poly_coefficients=self.poly_coefficients,
                relaxation_schedule=1,
                sum_constraint=100,
                solution_precision=1,
                lock_id="a bad lock",
            )
            self.assertDictEqual(result, JobCodes.LOCK_MISMATCH)
        finally:
            self.eqc_client.release_lock(lock_id=lock_id)

    def test_submit_job_precision_default(self):
        print("Test solve_sum_constrained precision default")
        lock_id, _, _ = self.eqc_client.wait_for_lock()
        try:
            result = self.eqc_client.solve_sum_constrained(
                poly_indices=self.poly_indices,
                poly_coefficients=self.poly_coefficients,
                relaxation_schedule=1,
                sum_constraint=2.4,
                lock_id=lock_id,
            )
        finally:
            self.eqc_client.release_lock(lock_id=lock_id)

    def test_solve_sum_constrained_precision_long_decimal(self):
        print("Test solve_sum_constrained precision long decimal")
        lock_id, _, _ = self.eqc_client.wait_for_lock()
        try:
            result = self.eqc_client.solve_sum_constrained(
                poly_indices=self.poly_indices,
                poly_coefficients=self.poly_coefficients,
                relaxation_schedule=1,
                sum_constraint=37.2515,
                solution_precision=0.372515,
                lock_id=lock_id,
            )
        finally:
            self.eqc_client.release_lock(lock_id=lock_id)

    def test_warn_soln_precision_exceeds_decimal(self):
        print("Test submit job precision_exceeds decimal")
        lock_id, _, _ = self.eqc_client.wait_for_lock()
        try:
            with self.assertWarns(Warning) as cm:
                result = self.eqc_client.submit_sum_constrained_job(
                    poly_indices=self.poly_indices,
                    poly_coefficients=self.poly_coefficients,
                    relaxation_schedule=1,
                    sum_constraint=500.25151278,
                    solution_precision=0.50025151278,
                    lock_id=lock_id,
                )
            self.assertEqual(2, len(cm.warnings))
            self.assertEqual(
                "`solution_precision`precision is greater than 7 decimal places. Will be modified on submission to device to float32 precision",
                str(cm.warnings[0].message),
            )
            self.assertEqual(
                "`sum_constraint` precision is greater than 7 decimal places. Will be modified on submission to device to float32",
                str(cm.warnings[1].message),
            )
        finally:
            self.eqc_client.release_lock(lock_id=lock_id)

    def test_no_warnings_solve_sum_constrained(self):
        print("test no warnings solve_sum_constrained")
        lock_id, _, _ = self.eqc_client.wait_for_lock()
        try:
            with warnings.catch_warnings(record=True) as warnings_log:
                result = self.eqc_client.solve_sum_constrained(
                    poly_indices=self.poly_indices,
                    poly_coefficients=self.poly_coefficients,
                    relaxation_schedule=1,
                    sum_constraint=500,
                    solution_precision=1,
                    lock_id=lock_id,
                )
            self.assertEqual([], warnings_log)
        finally:
            self.eqc_client.release_lock(lock_id=lock_id)

    def test_solve_integer_num_levels_out_of_range(self):
        print("Test test_solve_integer_num_levels_out_of_range")
        print("LEVELS", np.arange(self.num_variables, dtype=np.int32) + 1)
        try:
            lock_id, _, _ = self.eqc_client.wait_for_lock()
            with self.assertRaises(RuntimeError) as context:
                self.eqc_client.solve_integer(
                    poly_coefficients=self.poly_coefficients,
                    poly_indices=self.poly_indices,
                    lock_id=lock_id,
                    relaxation_schedule=1,
                    num_levels=np.arange(self.num_variables, dtype=np.int32) + 1,
                )

            self.assertEqual(
                str(context.exception),
                "Submission failed with response: {'err_code': 18, 'err_desc': "
                + "'All elements of input `num_levels` must be greater than 1'}",
            )
        finally:
            self.eqc_client.release_lock(lock_id=lock_id)

    def test_solve_integer(self):
        print("Test solve_integer:", np.arange(self.num_variables, dtype=np.int32) + 2)
        try:
            lock_id, _, _ = self.eqc_client.wait_for_lock()
            result = self.eqc_client.solve_integer(
                poly_coefficients=self.poly_coefficients,
                poly_indices=self.poly_indices,
                lock_id=lock_id,
                relaxation_schedule=1,
                num_levels=np.arange(self.num_variables, dtype=np.int32) + 2,
            )
            print(result)

            for solution in result["solution"]:
                self.assertTrue(
                    # Solutions should be integers with type float32. The integers should be
                    # small enough in magnitude so that they are still exactly represented.
                    all([val.is_integer() for val in solution])
                )
        finally:
            self.eqc_client.release_lock(lock_id=lock_id)

    def test_solve_integer_binary_optimization_via_scalar_num_levels(self):
        print("Test test_solve_integer_binary_optimization_via_scalar_num_levels")
        try:
            lock_id, _, _ = self.eqc_client.wait_for_lock()
            result = self.eqc_client.solve_integer(
                poly_coefficients=self.poly_coefficients,
                poly_indices=self.poly_indices,
                lock_id=lock_id,
                relaxation_schedule=1,
                num_levels=np.array(2, dtype=np.int32),  # Binary optimization.
            )
            print(result)
            self.assertTrue(
                # Solutions should be integers with type float32. The integers should be
                # small enough in magnitude so that they are still exactly represented.
                all([val in (0, 1) for val in result["solution"][0]])
            )
        finally:
            self.eqc_client.release_lock(lock_id=lock_id)

    def test_solve_integer_num_levels_raise_not_int(self):
        print("test num levels must be int")
        try:
            lock_id, _, _ = self.eqc_client.wait_for_lock()
            with self.assertRaises(ValueError) as context:
                self.eqc_client.solve_integer(
                    poly_coefficients=self.poly_coefficients,
                    poly_indices=self.poly_indices,
                    lock_id=lock_id,
                    relaxation_schedule=1,
                    num_levels=2.1,
                )

            self.assertEqual(str(context.exception), "`num_levels` must be type int")
        finally:
            self.eqc_client.release_lock(lock_id=lock_id)

    def test_integer_solve_levels_invalid(self):
        print("Test test_solve_integer levels invalid")
        try:
            lock_id, _, _ = self.eqc_client.wait_for_lock()
            with self.assertRaises(RuntimeError) as context:
                self.eqc_client.solve_integer(
                    poly_coefficients=self.poly_coefficients,
                    poly_indices=self.poly_indices,
                    lock_id=lock_id,
                    relaxation_schedule=1,
                    num_levels=1,
                    mean_photon_number=2,
                    quantum_fluctuation_coefficient=1,
                )

            self.assertEqual(
                str(context.exception),
                "Submission failed with response: {'err_code': 18, 'err_desc': "
                + "'All elements of input `num_levels` must be greater than 1'}",
            )
        finally:
            self.eqc_client.release_lock(lock_id=lock_id)

    def test_integer_solve_mean_photon_number_invalid(self):
        print("Test test_solve_integer mean_photon_number invalid")
        try:
            lock_id, _, _ = self.eqc_client.wait_for_lock()
            with self.assertRaises(RuntimeError) as context:
                self.eqc_client.solve_integer(
                    poly_coefficients=self.poly_coefficients,
                    poly_indices=self.poly_indices,
                    lock_id=lock_id,
                    relaxation_schedule=1,
                    mean_photon_number=2.01,
                    quantum_fluctuation_coefficient=1,
                    num_levels=np.arange(self.num_variables, dtype=np.int32) + 2,
                )

            self.assertEqual(
                str(context.exception),
                "Submission failed with response: {'err_code': 20, 'err_desc': "
                + "'Mean photon number if specified must be in range [0.0000667, 0.0066666]'}",
            )
        finally:
            self.eqc_client.release_lock(lock_id=lock_id)

    def test_integer_solve_mean_photon_number_invalid(self):
        print("Test test_solve_integer quantum_fluctuation_coefficient invalid")
        try:
            lock_id, _, _ = self.eqc_client.wait_for_lock()
            with self.assertRaises(RuntimeError) as context:
                self.eqc_client.solve_integer(
                    poly_coefficients=self.poly_coefficients,
                    poly_indices=self.poly_indices,
                    lock_id=lock_id,
                    num_levels=2,
                    relaxation_schedule=1,
                    mean_photon_number=2,
                    quantum_fluctuation_coefficient=51,
                )

            self.assertEqual(
                str(context.exception),
                "Submission failed with response: {'err_code': 21, 'err_desc': "
                + "Quantum fluctuation coefficient if specified must be in range [1, 50]'}",
            )
        finally:
            self.eqc_client.release_lock(lock_id=lock_id)

    def test_integer_solve_check_edge_mph_nlr(self):
        print("Test test_solve_integer quantum_fluctuation_coefficient invalid")
        try:
            lock_id, _, _ = self.eqc_client.wait_for_lock()
            x1 = self.eqc_client.solve_integer(
                poly_coefficients=self.poly_coefficients,
                poly_indices=self.poly_indices,
                lock_id=lock_id,
                relaxation_schedule=1,
                num_levels=3,
                mean_photon_number=0.0000666,
                quantum_fluctuation_coefficient=1,
            )
            print("INTEGER EDGE", x1)
            x2 = self.eqc_client.solve_integer(
                poly_coefficients=self.poly_coefficients,
                poly_indices=self.poly_indices,
                lock_id=lock_id,
                num_levels=4,
                relaxation_schedule=1,
                mean_photon_number=0.00666666,
                quantum_fluctuation_coefficient=50,
            )
        finally:
            self.eqc_client.release_lock(lock_id=lock_id)

    def test_integer_solve_mean_photon_number_invalid(self):
        print("Test test_solve_integer mean_photon_number invalid")
        try:
            lock_id, _, _ = self.eqc_client.wait_for_lock()
            with self.assertRaises(RuntimeError) as context:
                self.eqc_client.solve_integer(
                    poly_coefficients=self.poly_coefficients,
                    poly_indices=self.poly_indices,
                    lock_id=lock_id,
                    relaxation_schedule=1,
                    num_levels=2,
                    mean_photon_number=0.0066667,
                    quantum_fluctuation_coefficient=0.1,
                )

            self.assertEqual(
                str(context.exception),
                "Submission failed with response: {'err_code': 18, 'err_desc': "
                + "'All elements of input `num_levels` must be greater than 0'}",
            )
        finally:
            self.eqc_client.release_lock(lock_id=lock_id)

    def test_integer_solve_mean_photon_number_invalid(self):
        print("Test test_solve_integer mean_photon_number invalid")
        try:
            lock_id, _, _ = self.eqc_client.wait_for_lock()
            with self.assertRaises(RuntimeError) as context:
                self.eqc_client.solve_integer(
                    poly_coefficients=self.poly_coefficients,
                    poly_indices=self.poly_indices,
                    lock_id=lock_id,
                    relaxation_schedule=1,
                    mean_photon_number=0.0066668,
                    quantum_fluctuation_coefficient=1,
                    num_levels=[2] * self.num_variables,
                )

            self.assertEqual(
                str(context.exception),
                "Submission failed with response: {'err_code': 20, 'err_desc': "
                + "'Mean photon number if specified must be in range [0.0000667, 0.0066666]'}",
            )
        finally:
            self.eqc_client.release_lock(lock_id=lock_id)

    def test_sum_constrained_mean_qf_invalid(self):
        print("Test test_solve_integer quantum_fluctuation_coefficient invalid")
        try:
            lock_id, _, _ = self.eqc_client.wait_for_lock()
            with self.assertRaises(RuntimeError) as context:
                self.eqc_client.solve_sum_constrained(
                    poly_coefficients=self.poly_coefficients,
                    poly_indices=self.poly_indices,
                    lock_id=lock_id,
                    relaxation_schedule=2,
                    mean_photon_number=0.0066,
                    quantum_fluctuation_coefficient=101,
                )

            self.assertEqual(
                str(context.exception),
                "Submission failed with response: {'err_code': 21, 'err_desc': "
                + "'Quantum fluctuation coefficient if specified must be in range [1, 100]'}",
            )
        finally:
            self.eqc_client.release_lock(lock_id=lock_id)

    def test_sum_constrained_solve_check_edge_mph_nlr(self):
        print("Test test_solve_integer quantum_fluctuation_coefficient invalid")
        try:
            lock_id, _, _ = self.eqc_client.wait_for_lock()
            x1 = self.eqc_client.solve_sum_constrained(
                poly_coefficients=self.poly_coefficients,
                poly_indices=self.poly_indices,
                lock_id=lock_id,
                relaxation_schedule=1,
                mean_photon_number=0.0000667,
                quantum_fluctuation_coefficient=1,
            )
            print("MPN EDGE", x1)
            x2 = self.eqc_client.solve_sum_constrained(
                poly_coefficients=self.poly_coefficients,
                poly_indices=self.poly_indices,
                lock_id=lock_id,
                relaxation_schedule=1,
                mean_photon_number=0.0066666,
                quantum_fluctuation_coefficient=50,
            )
            print(x2)
        finally:
            self.eqc_client.release_lock(lock_id=lock_id)



import sys
import time
from multiprocessing import Pool
from eqc_direct.client import EqcClient
from eqc_direct.utils import convert_hamiltonian_to_poly_format
import numpy as np


def run_single():
    np.random.seed(13)

    ham_mat = np.loadtxt("QPLIB_0018_OBJ.csv", delimiter=",", dtype=float)
    poly_idx, poly_coef = convert_hamiltonian_to_poly_format(
        linear_terms=ham_mat[:, 0], quadratic_terms=ham_mat[:, 1:]
    )
    start = time.time()
    eqc_client = EqcClient()
    lock_id, start_ts, end_ts = eqc_client.wait_for_lock()
    print("Lock:", lock_id)
    resp = eqc_client.process_job(
        poly_indices=poly_idx,
        poly_coefficients=poly_coef,
        relaxation_schedule=1,
        sum_constraint=1,
        solution_precision=0.1,
        lock_id=lock_id,
    )
    eqc_client.release_lock(lock_id=lock_id)
    print(resp)


if __name__ == "__main__":
    run_single()
